from nirvana.lineage.abstractions import LineageNode


def rewire_nodes(start_node: LineageNode, node_list: list[LineageNode]):
    last_node = start_node
    for node in node_list:
        node.node_fields.left_input_fields = last_node.node_fields.output_fields
        node.node_fields.output_fields = list(
            set(node.node_fields.left_input_fields + node.operator.generated_fields)
        )
        node.set_left_child(last_node)
        last_node = node
    return last_node


class MapPullup:
    @classmethod
    def check_pattern(cls, dependencies: list[str], generated_fields: list[str]) -> bool:
        return all([field not in generated_fields for field in dependencies])

    @classmethod
    def split_maps_into_pullup_and_stay(cls, node: LineageNode, maps: list[LineageNode]) -> tuple[list, list]:
        maps_to_pullup, maps_to_stay = [], []
        for map in maps:
            dependencies = node.operator.dependencies
            map_generated_fields = map.operator.generated_fields
            can_pullup = cls.check_pattern(dependencies, map_generated_fields)
            if can_pullup:
                maps_to_pullup.append(map)
            else:
                maps_to_stay.append(map)
        return maps_to_pullup, maps_to_stay
    
    @classmethod
    def pull_up(cls, node: LineageNode) -> tuple[LineageNode, list[LineageNode]]:
        if node.op_name == "scan":
            return node, []

        elif node.op_name in ["filter", "rank"]:
            child_node, pullup_maps = cls.pull_up(node.left_child)
            maps_to_pullup, maps_to_stay = cls.split_maps_into_pullup_and_stay(node, pullup_maps)
            last_node = rewire_nodes(child_node, maps_to_stay)
            
            node.node_fields.left_input_fields = last_node.node_fields.output_fields
            node.node_fields.output_fields = node.node_fields.left_input_fields
            node.set_left_child(last_node)

            last_node = rewire_nodes(node, maps_to_pullup)
            return node, maps_to_pullup

        elif node.op_name == "join":
            left_child_node, left_pullup_maps = cls.pull_up(node.left_child)
            right_child_node, right_pullup_maps = cls.pull_up(node.right_child)

            left_maps_to_pullup, left_maps_to_stay = cls.split_maps_into_pullup_and_stay(node, left_pullup_maps)
            right_maps_to_pullup, right_maps_to_stay = cls.split_maps_into_pullup_and_stay(node, right_pullup_maps)

            left_last_node = rewire_nodes(left_child_node, left_maps_to_stay)
            right_last_node = rewire_nodes(right_child_node, right_maps_to_stay)
            
            node.node_fields.left_input_fields = left_last_node.node_fields.output_fields
            node.node_fields.right_input_fields = right_last_node.node_fields.output_fields

            maps_to_pullup = left_maps_to_pullup + right_maps_to_pullup
            last_node = rewire_nodes(node, maps_to_pullup)
            return node, maps_to_pullup

        elif node.op_name == "map":
            child_node, pullup_maps = cls.pull_up(node.left_child)
            maps_to_pullup, maps_to_stay = cls.split_maps_into_pullup_and_stay(node, pullup_maps)
            pullup_maps = maps_to_stay + [node] + maps_to_pullup
            last_node = rewire_nodes(child_node, pullup_maps)
            return child_node, pullup_maps

        elif node.op_name == "reduce":
            child_node, pullup_maps = cls.pull_up(node.left_child)
            last_node = rewire_nodes(child_node, pullup_maps)
            node.node_fields.left_input_fields = last_node.node_fields.output_fields
            node.set_left_child(last_node)
            return node, pullup_maps

    @classmethod
    async def transform(cls, node: LineageNode) -> tuple[LineageNode, float]:
        node, _ = cls.pull_up(node)
        return node, 0.0



import logging
from copy import deepcopy
from pathlib import Path
from typing import Union

from nirvana.executors.llm_backbone import LLMClient
from nirvana.ops.base import BaseOperation
from nirvana.optim.optimizer import PlanOptimizer


def configure_llm_backbone(
    model_name: str = None, api_key: Union[str, Path] = None, **kwargs
):
    """
    Configures the LLM (Large Language Model) backbone by initializing and setting it 
    for use in the application.
    Args:
        model_name (str, optional): The name of the LLM model to be used. Defaults to None.
        api_key (Union[str, Path], optional): The API key or path to the API key required 
            to authenticate with the LLM service. Defaults to None.
        base_url (str, optional): The base URL of the LLM service endpoint. Defaults to None.
    Returns:
        None

    Example:
        ```python
        >>> import mahjong as mjg
        >>> mjg.configure_llm_backbone(model_name="gpt-3.5-turbo", api_key="your-api-key")
        ```
    """
    llm_client = LLMClient.configure(model_name, api_key, **kwargs)
    BaseOperation.set_llm(deepcopy(llm_client))
    PlanOptimizer.set_agent(deepcopy(llm_client))



import pandas as pd
import nirvana as nvn


def test_image_data():
    nvn.configure_llm_backbone(
        model_name="Deepseek-chat",
        api_key="<Your-API-Key>",
        base_url="https://api.deepseek.com"
    )

    logo_imgs = nvn.ImageArray([
        "https://spark.apache.org/images/spark-logo.png",
        "https://upload.wikimedia.org/wikipedia/commons/thumb/c/c6/PyTorch_logo_black.svg/488px-PyTorch_logo_black.svg.png",
        "https://upload.wikimedia.org/wikipedia/commons/thumb/6/63/Databricks_Logo.png/960px-Databricks_Logo.png"
    ])
    df = pd.DataFrame({
        "names": ["Spark", "Pytorch", "Databricks"],
        "logos": logo_imgs
    })
    
    outputs = nvn.ops.filter(df, "Whether the image is a software logo?", input_schema="logos", strategy="plain_llm")
    return outputs



import os
import multiprocessing
from multiprocessing import Pool
import pandas as pd

from nirvana.dataframe.arrays.image import ImageDtype, ImageArray, load_image
from nirvana.dataframe.arrays.audio import AudioDtype, AudioArray, load_audio
from nirvana.dataframe.arrays.file import FileDtype, FileArray, load_file


EXT_TYPE_MAPPING = {
    "image": [".jpg", ".jpeg", ".png", ".bmp", ".gif", ".tiff", ".webp", ".svg", ".ico", ".jfif"],
    "audio": [".mp3", ".wav", ".ogg", ".flac", ".aac", ".m4a", ".wma", ".opus"],
    "video": [".mp4", ".mov", ".avi", ".mkv", ".webm", ".flv", ".wmv"],
    "pdf": [".pdf"],
    "doc": [".doc", ".docx", ".txt", ".rtf", ".odt"],
    "ppt": [".ppt", ".pptx", ".odp"],
    # "excel": [".xls", ".xlsx", ".csv"],
}


def infer_dtype(col: pd.Series):
    NUM_SAMPLE = 10
    file_path = col.sample(NUM_SAMPLE)
    ext = file_path.str.extract(r"(?P<ext>\.[^\.]+(\.gz)?)$")["ext"].str.lower()

    for type, exts in EXT_TYPE_MAPPING.items():
        if ext.isin(exts).any():
            return type
    return "text"


def infer_and_convert_dtype(col: pd.Series) -> tuple[pd.Series, type]:
    """
    Infer and return a specific Python or extension dtype/type (e.g., int, float, str, ImageArray, AudioArray, etc.)
    for a given column-like structure. Distinguish between pandas/pyarrow types and unstructured
    data, such as image/audio file URLs and paths.
    """
    if col.empty:
        return col, col.dtype
    
    dtype = col.dtype
    num_workers = multiprocessing.cpu_count()
    if pd.api.types.is_string_dtype(dtype):
        inferred_dtype = infer_dtype(col)
        if inferred_dtype == "image":
            with Pool(num_workers) as pool:
                col = pool.map(load_image, col.values)
            return ImageArray(col), ImageDtype
        elif inferred_dtype == "audio":
            with Pool(num_workers) as pool:
                col = pool.map(load_audio, col.values)
            return AudioArray(col), AudioDtype
        elif inferred_dtype == "pdf":
            with Pool(num_workers) as pool:
                col = pool.map(load_file, col.values)
            return FileArray(col), FileDtype
        else:
            return col, str
    else:
        return col, dtype



from dataclasses import asdict
from nirvana.lineage.abstractions import LineageNode


class FilterPushdown:
    @classmethod
    def check_pattern(cls, dependencies: list[str], existing_fields: list[str]) -> bool:
        return all([field in existing_fields for field in dependencies])

    @classmethod
    async def push_down(cls, node: LineageNode) -> LineageNode:
        if node.op_name == "filter":
            last_node = await cls.push_down(node.left_child)
            input_columns = node.operator.input_columns

            if last_node.op_name == "join":
                left_fields = last_node.node_fields.left_input_fields
                right_fields = last_node.node_fields.right_input_fields
                # push filter into the left sub-lineage
                pushdown_flag = False
                if cls.check_pattern(input_columns, left_fields):
                    new_node = LineageNode(op_name="filter", op_kwargs=node.operator.op_kwargs, node_fields=asdict(node.node_fields))
                    # swap info (eg fields) of current op (eg, filter) and its predecessor (ie join)
                    new_node.node_fields.output_fields = new_node.node_fields.left_input_fields = last_node.node_fields.left_input_fields
                    # rewire edges between current op and its predecessor and rewrite sub-lineage over pushdowned filter
                    new_node.set_left_child(last_node.left_child)
                    last_node.set_left_child(await cls.push_down(new_node))
                    pushdown_flag = True
                # push filter into the right sub-lineage
                if cls.check_pattern(input_columns, right_fields):
                    new_node = LineageNode(op_name="filter", op_kwargs=node.operator.op_kwargs, node_fields=asdict(node.node_fields))
                    # swap info (eg fields) of current op, filter, and its predecessor (ie join)
                    new_node.node_fields.output_fields = new_node.node_fields.left_input_fields = last_node.node_fields.right_input_fields
                    # rewire edges between current op and its predecessor and rewrite sub-lineage over pushdowned filter
                    new_node.set_left_child(last_node.right_child)
                    last_node.set_right_child(await cls.push_down(new_node))
                    pushdown_flag = True
                if pushdown_flag:
                    del node
                    return last_node
                else:
                    node.set_left_child(last_node)
                    return node
            
            elif last_node.op_name in ["map", "filter", "rank"]:
                fields = last_node.node_fields.left_input_fields
                if cls.check_pattern(input_columns, fields):
                    # swap info (eg fields) of current op, filter, and its predecessor (eg map)
                    node.node_fields.output_fields = node.node_fields.left_input_fields = last_node.node_fields.left_input_fields
                    # rewire edges around current op and its predecessor, and rewrite sub-lineage over pushdowned filter
                    node.set_left_child(last_node.left_child)
                    last_node.set_left_child(await cls.push_down(node))
                    return last_node
                else:
                    node.set_left_child(last_node)
                    return node
                
            else:
                node.set_left_child(last_node)
                return node
                
        elif node.op_name == "join":
            node.set_left_child(await cls.push_down(node.left_child))
            node.set_right_child(await cls.push_down(node.right_child))
            return node
        
        elif node.op_name in ["map", "rank", "reduce"]:
            node.set_left_child(await cls.push_down(node.left_child))
            return node
        
        else:
            return node

    @classmethod
    async def transform(cls, node: LineageNode) -> tuple[LineageNode, float]:
        node = await cls.push_down(node)
        return node, 0.0



import pandas as pd


class JoinPrompter:
    def __init__(self):
        self.task_instruction: str = None
        self.output_format: str = None

    def prepare_prompt_for_nested_join(self):
        self.task_instruction = (
            "You are a helpful assistant helping the user make sense of their data. "
            "You are performing a join operation (left data and right data -> True or False) to "
            "determine whether the two given data satisfy the given conditions."
        )
        self.output_format = (
            "Output the result of the join operation concisely in the following format.\n"
            "<output> Your final answer from [True, False] </output>\n"
        )

    def generate_nested_join_prompt(
        self, 
        left_data: pd.Series,
        right_data: pd.Series,
        user_instruction: str | list[str],
        left_dtypes: list[str],
        right_dtypes: list[str],
    ):
        # 1. Prepare system message
        if self.task_instruction is None or self.output_format is None:
            self.prepare_prompt_for_nested_join()
        sys_message = [
            {"role": "system", "content": self.task_instruction},
            {"role": "system", "content": self.output_format}
        ]

        # 2. Prepare data
        user_content = []
        user_content.append({"type": "input_text", "text": "Left data:"})
        for dtype, (key, value) in zip(left_dtypes, left_data.items()):
            if dtype == "text":
                user_content.append({"type": "input_text", "text": f"{key}: {value}"})
            elif dtype == "image":
                user_content.append({"type": "input_text", "text": f"{key}:"})
                user_content.append({"type": "input_image", "image_url": value})
            elif dtype == "audio":
                user_content.append({"type": "input_text", "text": f"{key}:"})
                user_content.append(
                    {"type": "input_audio", "input_audio": {"data": value, "format": "wav"}}
                )
            else:
                raise ValueError(f"Data type of left data {type(value)} is not supported.")
        
        user_content.append({"type": "input_text", "text": "Right data:"})
        for dtype, (key, value) in zip(right_dtypes, right_data.items()):
            if dtype == "text":
                user_content.append({"type": "input_text", "text": f"{key}: {value}"})
            elif dtype == "image":
                user_content.append({"type": "input_text", "text": f"{key}:"})
                user_content.append({"type": "input_image", "image_url": value})
            elif dtype == "audio":
                user_content.append({"type": "input_text", "text": f"{key}:"})
                user_content.append(
                    {"type": "input_audio", "input_audio": {"data": value, "format": "wav"}}
                )
            else:
                raise ValueError(f"Data type of right data {type(value)} is not supported.")
        
        # 3. Prepare the given condition
        if isinstance(user_instruction, str):
            conditions = f"condition: {user_instruction}"
        elif isinstance(user_instruction, list):
            conditions = [f"condition {idx}: {cond}" for idx, cond in enumerate(user_instruction)]
            conditions = "\n".join(conditions)
        user_content.append({"type": "input_text", "text": conditions})
        
        user_message = [{"role": "user", "content": user_content}]

        messages = sys_message + user_message
        return messages
    
    def prepare_prompt_for_block_join(self):
        self.task_instruction = (
            "Identify pairs of items from the left and right data batches that satisfy the given join conditions."
        )
        self.output_format = (
            "Output only the IDs of pairs satisfying the join conditions in the format of L#-R# (e.g., L3-R5), separated by commas. ",
            "Make sure that all candidate pairs of items are considered.\n",
            "The result of the join operation concisely in the following format.\n",
            "<output> L3-R5,L4-R2,L1-R1 </output>\n"
            "If no pairs satisfy the conditions, output an empty <output></output> tag.\n"
        )
    
    def generate_batch_join_prompt(
        self,
        left_batch: pd.DataFrame,
        right_batch: pd.DataFrame,
        user_instruction: str | list[str],
        left_dtypes: list[str],
        right_dtypes: list[str],
    ):
        # 1. Prepare system message
        if self.task_instruction is None or self.output_format is None:
            self.prepare_prompt_for_block_join()
        sys_message = [
            {"role": "system", "content": self.task_instruction},
            {"role": "system", "content": self.output_format}
        ]

        # 2. Prepare data
        user_content = []
        # 2.1 Prepare left batch
        user_content.append({"type": "input_text", "text": "# Left batch:"})
        for batch_idx, left_data in left_batch.iterrows():
            user_content.append({"type": "input_text", "text": f"## L{batch_idx}:"})
            for dtype, (key, value) in zip(left_dtypes, left_data.items()):
                if dtype == "text":
                    user_content.append({"type": "input_text", "text": f"{key}: {value}"})
                elif dtype == "image":
                    user_content.append({"type": "input_text", "text": f"{key}:"})
                    user_content.append({"type": "input_image", "image_url": value})
                elif dtype == "audio":
                    user_content.append({"type": "input_text", "text": f"{key}:"})
                    user_content.append(
                        {"type": "input_audio", "input_audio": {"data": value, "format": "wav"}}
                    )
                else:
                    raise ValueError(f"Data type of left data {type(value)} is not supported.")
        
        # 2.2 Prepare right batch
        user_content.append({"type": "input_text", "text": "# Right batch:"})
        for batch_idx, right_data in right_batch.iterrows():
            user_content.append({"type": "input_text", "text": f"## R{batch_idx}:"})
            for dtype, (key, value) in zip(right_dtypes, right_data.items()):
                if dtype == "text":
                    user_content.append({"type": "input_text", "text": f"{key}: {value}"})
                elif dtype == "image":
                    user_content.append({"type": "input_text", "text": f"{key}:"})
                    user_content.append({"type": "input_image", "image_url": value})
                elif dtype == "audio":
                    user_content.append({"type": "input_text", "text": f"{key}:"})
                    user_content.append(
                        {"type": "input_audio", "input_audio": {"data": value, "format": "wav"}}
                    )
                else:
                    raise ValueError(f"Data type of right data {type(value)} is not supported.")
        
        # 3. Prepare the given condition
        if isinstance(user_instruction, str):
            conditions = f"condition: {user_instruction}"
        elif isinstance(user_instruction, list):
            conditions = [f"condition {idx}: {cond}" for idx, cond in enumerate(user_instruction)]
            conditions = "\n".join(conditions)
        user_content.append({"type": "input_text", "text": conditions})
        
        user_message = [{"role": "user", "content": user_content}]
        messages = sys_message + user_message
        return messages



from nirvana.executors.llm_backbone import LLMArguments, LLMClient
from nirvana.executors.tools import FunctionCallTool, PythonInterpreterTool

__all__ = [
    # llms
    "LLMArguments",
    "LLMClient",
    # tools
    "FunctionCallTool",
    "PythonInterpreterTool",
]



from typing import Any, Iterable


class RankPrompter:
    
    def __init__(self):
        self.system_instruction = (
            "Your task is to select the most relevant item from the given two data to the user-specified criterion. "
            "Responde only with the label of the data such as 'item number', i.e., "
            "the response must be either 1 or 2 (things like 'None' or 'Neither' are forbidden), "
            "depending on which item is more relevant to the criterion.\n"
            "Output the item number concisely in the following format.\n"
            "<output> item number (1 or 2) </output>\n"
        )

    def generate_prompt(
        self, 
        data1: Any,
        data2: Any,
        user_instruction: str,
        dtype: str = "text",
    ):
        # 1. Prepare system message
        sys_message = [{"role": "system", "content": self.system_instruction}]

        # 2. Prepare user message
        user_content = []
        if dtype == "text":
            user_content.append({"type": "input_text", "text": f"Item 1:\n{str(data1)}"})
            user_content.append({"type": "input_text", "text": f"Item 2:\n{str(data2)}"})
        elif dtype == "image":
            user_content.append({"type": "input_text", "text": f"Item 1:"})
            user_content.append({"type": "input_image", "image_url": data1})
            user_content.append({"type": "input_text", "text": f"Item 2:"})
            user_content.append({"type": "input_image", "image_url": data2})
        else:
            raise ValueError(f"Data type {dtype} is not supported.")
        user_content.append({"type": "input_text", "text": f"Criterion: {user_instruction}"})
        user_message = [{"role": "user", "content": user_content}]

        messages = sys_message + user_message
        return messages



import re
import asyncio
from collections import deque

from nirvana.executors.llm_backbone import LLMClient
from nirvana.executors.tools import FunctionCallTool
from nirvana.lineage.abstractions import LineageNode


def build_code_from_lineage(last_node_in_plan: LineageNode) -> str:
    expressions = []
    datasets_info = []
    def _build_expression(node: LineageNode):
        if node.op_name == "scan":
            datasets_info.append(
                f"- Dataset {len(datasets_info) + 1}: df{len(datasets_info) + 1} with columns {node.node_fields.output_fields}"
            )
            return
        
        _build_expression(node.left_child)
        if node.right_child:
            _build_expression(node.right_child)

        if node.op_name == "filter":
            expression = (
                f"df{len(datasets_info)}.semantic_{node.op_name}(user_instruction=\"{node.operator.user_instruction}\", input_columns={node.operator.input_columns})"
            )
        elif node.op_name in {"rank", "reduce"}:
            expression = (
                f"df{len(datasets_info)}.semantic_{node.op_name}(user_instruction=\"{node.operator.user_instruction}\", input_column=\"{node.operator.input_columns[0]}\")"
            )
        elif node.op_name == "map":
            expression = (
                f"df{len(datasets_info)}.semantic_{node.op_name}(user_instruction=\"{node.operator.user_instruction}\", input_columns={node.operator.input_columns}, output_columns={node.operator.output_columns})"
            )
        elif node.op_name == "join":
            expression = (
                f"df{len(datasets_info)}.semantic_{node.op_name}(other=df{len(datasets_info) + 2}, user_instruction=\"{node.operator.user_instruction}\", left_on=\"{node.operator.left_on[0]}\", right_on=\"{node.operator.right_on[0]}\", how=\"{node.operator.how}\")"
            )
        else:
            raise ValueError(f"Unsupported operation {node.op_name} for code generation.")
        expressions.append(expression)
        return

    _build_expression(last_node_in_plan)
    code = "\n".join(expressions)
    dataset_info = "\n".join(datasets_info)
    return code.strip(), dataset_info


def extract_udfs_from_code(code: str) -> deque:
    udfs = deque()
    expressions = code.split("\n")
    for expr in expressions:
        match = re.search(r'\w+\.semantic_(\w+)\((.*)\)', expr, flags=re.DOTALL)
        if match:
            op_name = match.group(1)
            args_str = match.group(2)
            udf_str = re.search(r'func=([^,]+)', args_str)
            if udf_str:
                udf_code = udf_str.group(1).strip()
                try:
                    udfs.append((udf_code, eval(udf_code)))
                except Exception as e:
                    raise ValueError(f"Failed to evaluate UDF: {udf_code}") from e
            else:
                udfs.append(None)
    return udfs


def replace_with_udf_in_lineage(node: LineageNode, udfs: deque) -> LineageNode:
    def _replace_in_node(node: LineageNode):
        if node.op_name == "scan":
            return
        _replace_in_node(node.left_child)
        if node.right_child:
            _replace_in_node(node.right_child)

        udf = udfs.popleft()
        if udf:
            node.operator.tool = FunctionCallTool.from_function(
                name=udf[0], func=udf[1],
            )
        else:
            return
    
    _replace_in_node(node)
    return node


class NonLLMReplace:
    """
    Replace LLM-driven evaluations on NL instructions with equivalent UDFs.
    """

    rewrite_prompt = """Given a dataset and a user-specified query for semantic data processing.
Each dataset is given in the form of a pandas-like dataframe, and the query is represented by a sequence of .semantic_*(...) operators (* can be replaced with names of supported operators).

The supported operators and their required arguments are as follows.
1. map: Perform an element-wise projection specified by natural language on a given column to a new column. Required arguments:
- user_instruction: a natural language expression
- input_columns: the names of the columns on which the operation is performed
- output_columns: the field names that the operation generates
- func: a lambda function applied to each row of DataFrame (i.e., pd.Series) against input columns, returns a dict with output_column as keys and converted values as values
2. filter: Evaluate a natural language condition per value in a given column (returning boolean). Required arguments:
- user_instruction: the natural language condition
- input_columns: the names of columns on which the operation is performed
- func: a lambda function applied to each row of DataFrame (i.e., pd.Series) against input columns, returns a boolean value per value
3. join: Join a table with another table by keeping all tuple pairs that satisfy a natural language condition. Required arguments:
- other: the other dataset to join with
- user_instruction: the join condition in natural language
- left_on: the name of the column from the left table to join on
- right_on: the name of the column from the right table to join on
- how: the type of join to be performed (e.g., inner, left, right)
- func: a lambda function applied to two rows from left table and right table (i.e., two pd.Series) against left_on and right_on, respectively, returns a boolean value per tuple pair
4. reduce: Aggregate multiple values in a given column into a single result. Required arguments:
- user_instruction: the reducer function in natural language
- input_column: the name of column on which the operation is performed
- func: a lambda function processed by underlying pd.DataFrame.agg() for aggregating the data

Here is an example of a semantic data processing query that contains only map and filter operations:
```python
df.semantic_map(user_instruction="map instruction", input_columns=["col_a"], output_columns=["col_b"])
df.semantic_filter(user_instruction="filter instruction", input_columns=["col_c"])
```

Now, you are given following dataset(s):
{dataset_info}
and a data processing workflow as follows:
```python
{logical_plan}
```

You are tasked with replacing the NL instruction with an equivalent compute function for as many operations as possible. The rewrite aims to generate a plan that is semantically equivalent to the original plan but reduces LLM calls.
There are several constraints to follow.
- The replacement is applied only when `user_instruction` can be converted to a built-in function or a lambda expression.
- If no appropriate replacement applied, keep the original operation.
- Except adding argument `func`, do not change the data processing workflow.
- Any modification to the pre-defined operator interfaces is not allowed.
The rewrite is output as executable python code. Note that **every single operation is placed in a line, i.e., no line break for arguments**. If no rewrites proposed, return an empty code block. ONLY ONE code block can be contained in the output.
"""

    @classmethod
    async def apply(cls, node: LineageNode, rewriter: LLMClient) -> tuple[LineageNode, float]:
        code, dataset_info = build_code_from_lineage(node)
        if not code:
            return node, 0.0

        prompt = cls.rewrite_prompt.format(
            dataset_info=dataset_info,
            logical_plan=code
        )
        response = await rewriter(prompt, parse_code=True, lang="python")
        code, rewrite_cost = response["output"], response["cost"]

        udfs = extract_udfs_from_code(code)
        if udfs:
            new_plan = replace_with_udf_in_lineage(node, udfs)
            return new_plan, rewrite_cost
        else:
            return node, rewrite_cost

    @classmethod
    async def transform(cls, node: LineageNode, rewriter: LLMClient) -> tuple[LineageNode, float]:
        return await cls.apply(node, rewriter=rewriter)



import pandas as pd
from typing import Any, List, Dict


class MapPrompter:
    def __init__(self):
        self.system_instruction = (
            "You are a helpful assistant helping the user make sense of their data. "
            "You are performing a map operation to project the given data into correct values for a set of output fields based on the user's instruction."
        )
        self.output_format: str = None

    def prepara_output_format(self, output_columns: list[str]):
        output_format = "Output the result of the map operation (each field) concisely in the following format.\n"
        for column_name in output_columns:
            output_format += f"<{column_name}> fill in correct value </{column_name}>\n"
        return output_format

    def generate_prompt(
        self, 
        data: pd.Series,
        user_instruction: str,
        output_columns: list[str],
        dtypes: list[str],
    ):
        """
        Generates a prompt message for LLMs based on user instructions 
        and provided data.

        Args:
            user_instruction (str): The instruction or query provided by the user.
            data (Any): The data to be included in the prompt. It can be a string or an 
                instance of `mjg.ImageDtype`. If the data type is unsupported, a ValueError 
                is raised.

        Returns:
            list: A list of dictionaries representing the structured prompt messages. 
            The messages include a system message and a user message, where the user 
            message contains the provided data and user instruction.

        Raises:
            ValueError: If the type of `data` is not supported.
        """
        # 1. Prepare system message
        if self.output_format is None:
            self.output_format = self.prepara_output_format(output_columns=output_columns)
        sys_message = [
            {"role": "system", "content": self.system_instruction},
            {"role": "system", "content": self.output_format}
        ]

        # 2. Prepare data
        user_content = []
        for dtype, (key, val) in zip(dtypes, data.items()):
            if dtype == "text":
                user_content.append({"type": "input_text", "text": f"{key}: {val}"})
            elif dtype == "image":
                user_content.append({"type": "input_text", "text": f"{key}:"})
                user_content.append({"type": "input_image", "image_url": val})
            elif dtype == "audio":
                user_content.append({"type": "input_text", "text": f"{key}:"})
                user_content.append(
                    {"type": "input_audio", "input_audio": {"data": val, "format": "wav"}}
                )
            else:
                raise ValueError(f"Data type {dtype} is not supported.")
        
        # 3. Prepare the given transformation instruction
        user_content.append({"type": "input_text", "text": user_instruction})
        user_message = [{"role": "user", "content": user_content}]

        messages = sys_message + user_message
        return messages
    
    def generate_fewshot_prompt(
        self,
        data: pd.Series,
        user_instruction: str,
        output_columns: list[str],
        dtypes: list[str],
        demos: list[dict[str, Any]]
    ):
        """
        Generates a prompt with demonstration examples for an LLM.
        Args:
            user_instruction (str): The instruction provided by the user to guide the AI's response.
            data (Any): The input data provided by the user. Must be a string; otherwise, a ValueError is raised.
            demos (List[Dict[str, Any]]): A list of demonstration examples. Each example should be a dictionary 
                containing the keys "data" (input data as a string) and "answer" (expected output as a string). 
                If a demo is not a string, a ValueError is raised.
        Returns:
            List[Dict[str, Any]]: A list of messages formatted for the conversational AI system. The messages 
            include:
                - A system message containing the system's instruction.
                - Demonstration messages based on the provided examples.
                - A user message containing the input data and user instruction.
        Raises:
            ValueError: If the type of `data` or `demo["data"]` is not supported.
        """
        # 1. Prepare system message
        if self.output_format is None:
            self.output_format = self.prepara_output_format(output_columns=output_columns)
        sys_message = [
            {"role": "system", "content": self.system_instruction},
            {"role": "system", "content": self.output_format}
        ]

        # 2. Prepare demonstration message
        demos_message = [{"type": "input_text", "text": f"Here are some examples:"}]
        for demo in demos:
            demo_content = []
            demo_data: pd.Series | dict = demo["data"]
            demo_answer: str = demo["answer"]
            for dtype, (key, val) in zip(dtypes, demo_data.items()):
                if dtype == "text":
                    demo_content.append({"type": "input_text", "text": f"{key}: {val}"})
                elif dtype == "image":
                    demo_content.append({"type": "input_text", "text": f"{key}:"})
                    demo_content.append({"type": "input_image", "image_url": val})
                elif dtype == "audio":
                    demo_content.append({"type": "input_text", "text": f"{key}:"})
                    demo_content.append(
                        {"type": "input_audio", "input_audio": {"data": val, "format": "wav"}}
                    )
                else:
                    raise ValueError(f"Data type {dtype} is not supported.")
            demo_content.append({"type": "input_text", "text": f"condition: {user_instruction}"})
            demo_content.append({"type": "input_text", "text": f"Answer: {demo_answer}"})
            demos_message.append(
                {"role": "assistant", "content": demo_content}
            )

        # 3. Prepare user message
        user_content = []
        for dtype, (key, val) in zip(dtypes, data.items()):
            if dtype == "text":
                user_content.append({"type": "input_text", "text": f"{key}: {val}"})
            elif dtype == "image":
                user_content.append({"type": "input_text", "text": f"{key}:"})
                user_content.append({"type": "input_image", "image_url": val})
            elif dtype == "audio":
                user_content.append({"type": "input_text", "text": f"{key}:"})
                user_content.append(
                    {"type": "input_audio", "input_audio": {"data": val, "format": "wav"}}
                )
            else:
                raise ValueError(f"Data type {dtype} is not supported.")
        
        # 4. Prepare the given condition
        user_content.append({"type": "input_text", "text": user_instruction})
        user_message = [{"role": "user", "content": user_content}]
        
        messages = sys_message + demos_message + user_message
        return messages
    
    def generate_evaluate_prompt(
        self,
        data: Any,
        answer: Any,
        user_instruction: str,
        dtypes: list[str],
    ):
        evaluator_task = [{"role": "user", "content": "Analyze the map operation tasked with transforming data based on a given instruction:"}]
        
        _input = [{"type": "input_text", "text": "Input:"}]
        for dtype, (key, val) in zip(dtypes, data.items()):
            if dtype == "text":
                _input.append({"type": "input_text", "text": f"{key}: {val}"})
            elif dtype == "image":
                _input.append({"type": "input_text", "text": f"{key}:"})
                _input.append({"type": "input_image", "image_url": val})
            elif dtype == "audio":
                _input.append({"type": "input_text", "text": f"{key}:"})
                _input.append(
                    {"type": "input_audio", "input_audio": {"data": val, "format": "wav"}}
                )
            else:
                raise ValueError(f"Data type {dtype} is not supported.")
        
        _output = [{"type": "input_text", "text": f"Output:\n{answer}"}]

        instruction = user_instruction.strip() + " " + self.output_format
        _instruction = [{"type": "input_text", "text": f"Instruction:\n{instruction}"}]
        
        evaluator_context = [
            {"role": "user", "content": _input + _output + _instruction}
        ]

        evaluator_criteria = (
            "Evaluate the map operation based on the following criteria:\n"
            "1. Does the output strictly adhere to the required output format?\n"
            "2. Does the output satisfy the instruction?\n"
            "You should be evaluating only and not attemping to solve the task. Only output PASS if all criteria are met and you have no further suggestions for improvements.\n"
            "Output your evaluation concisely in the following format.\n"
            "<evaluation> PASS or FAIL </evaluation>\n"
            "<feedback> What needs improvement and why. </feedback>\n"
        )
        evaluator_criteria = [{"role": "user", "content": evaluator_criteria}]
        messages = evaluator_task + evaluator_context + evaluator_criteria
        return messages
    
    def generate_refine_prompt(
        self,
        data: Any,
        answer: Any,
        user_instruction: str,
        output_columns: list[str],
        feedback: str,
        dtypes: list[str],
    ):
        initial_generate_prompt = self.generate_prompt(data, user_instruction, output_columns, dtypes)
        refine_context =(
            "There is feedback from your previous output.\n"
            f"Output:\n{answer}\nFeedback:\n{feedback}\n"
            "Your task is to refine the output based on the feedback."
        )
        refine_prompt = [{"role": "user", "content": refine_context}]
        messages = initial_generate_prompt + refine_prompt
        return messages



from nirvana.dataframe.arrays.image import ImageDtype
from nirvana.dataframe.arrays.audio import AudioDtype
from nirvana.dataframe.arrays.file import FileDtype


__all__ = [
    "ImageDtype",
    "AudioDtype",
    "FileDtype",
]



import warnings

import functools
import asyncio
from typing import Any, Iterable, Callable, Literal
from dataclasses import dataclass, field
import pandas as pd

from nirvana.dataframe.arrays.image import ImageDtype
from nirvana.executors.tools import BaseTool, FunctionCallTool
from nirvana.ops.base import BaseOpOutputs, BaseOperation
from nirvana.ops.prompt_templates.map_prompter import MapPrompter


def map_wrapper(
    input_data: pd.DataFrame, 
    user_instruction: str = None,
    input_columns: list[str] = None,
    output_columns: list[str] = None,
    context: list[dict] | str | None = None,
    model: str | None = None,
    func: Callable = None,
    strategy: Literal["plain", "fewshot", "self-refine"] = "plain",
    limit: int | None = None,
    rate_limit: int = 16,
    assertions: list[Callable] | None = [],
    **kwargs
):
    """
    A function wrapper for map operation

    Args:
        input_data (pd.DataFrame): Input dataframe
        user_instruction (str, optional): User instruction. Defaults to None.
        input_columns (list[str], optional): Input column. Defaults to None.
        output_columns (list[str], optional): Output columns. Defaults to None.
        context (list[dict] | str, optional): Context. Defaults to None.
        model (str, optional): Model. Defaults to None.
        func (Callable, optional): User function. Defaults to None.
        strategy (Literal["plain", "fewshot", "self-refine"], optional): Strategy. Defaults to "plain".
        limit (int): Maximum number of outputs to produce before stopping.
        rate_limit (int, optional): Rate limit. Defaults to 16.
        assertions (list[Callable], optional): Assertions. Defaults to [].
        **kwargs: Additional keyword arguments for OpenAI Clent.
    """
    
    map_op = MapOperation(
        user_instruction=user_instruction,
        strategy=strategy,
        input_columns=input_columns,
        output_columns=output_columns,
        context=context,
        model=model,
        tool=FunctionCallTool.from_function(func=func) if func else None,
        limit=limit,
        rate_limit=rate_limit,
        assertions=assertions
    )
    outputs = asyncio.run(map_op.execute(
        input_data=input_data,
        **kwargs
    ))
    return outputs


@dataclass
class MapOpOutputs(BaseOpOutputs):
    outputs: dict[str, list] = field(default_factory=dict)

    def __add__(self, other: "MapOpOutputs"):
        assert self.outputs.keys() == other.outputs.keys(), "Cannot merge MapOpOutputs with different field names."
        new_output = {}
        for field_name in self.outputs.keys():
            new_output[field_name] = self.outputs[field_name] + other.outputs[field_name]
        return MapOpOutputs(
            outputs=new_output,
            cost=self.cost + other.cost
        )


class MapOperation(BaseOperation):
    """
    Map operator: Applies an LLM to perform a transformation on a column, producing a new column as output
    """
    strategy_options = ["plain", "fewshot", "self-refine"]
    
    def __init__(
        self,
        user_instruction: str = "",
        input_columns: list[str] = [],
        output_columns: list[str] = [],
        context: list[dict] | str | None = None,
        model: str | None = None,
        tool: Callable | BaseTool | None = None,
        strategy: Literal["plain", "fewshot", "self-refine"] = "plain",
        limit: int | None = None,
        rate_limit: int = 16,
        assertions: list[Callable] | None = [],
    ):
        if tool and not isinstance(tool, BaseTool):
            tool = FunctionCallTool.from_function(func=tool)
        
        super().__init__(
            op_name="map",
            user_instruction=user_instruction,
            context=context,
            model=model,
            tool=tool,
            strategy=strategy,
            limit=limit,
            rate_limit=rate_limit,
            assertions=assertions,
        )
        self.prompter = MapPrompter()
        self.input_columns = input_columns
        self.output_columns = output_columns

    @property
    def dependencies(self) -> list[str]:
        return self.input_columns
    
    @property
    def generated_fields(self) -> list[str]:
        return self.output_columns
    
    @property
    def op_kwargs(self):
        kwargs = super().op_kwargs
        kwargs["input_columns"] = self.input_columns
        kwargs["output_columns"] = self.output_columns
        return kwargs

    async def _execute_by_plain_llm(self, data: pd.Series, user_instruction: str, dtypes: list[str], **kwargs):
        async with self.semaphore:
            full_prompt = self.prompter.generate_prompt(data, user_instruction, self.output_columns, dtypes)
            output = await self.llm(full_prompt, parse_tags=True, tags=self.output_columns, **kwargs)
            result = self._postprocess_map_output(output, self.output_columns)
            return result, output["cost"]

    async def _execute_by_fewshot_llm(self, data: pd.Series, user_instruction: str, dtypes: list[str], demos, **kwargs):
        async with self.semaphore:
            full_prompt = self.prompter.generate_fewshot_prompt(data, user_instruction, self.output_columns, dtypes, demos)
            output = await self.llm(full_prompt, parse_tags=True, tags=self.output_columns, **kwargs)
            result = self._postprocess_map_output(output, self.output_columns)
            return result, output["cost"]
        
    async def _execute_by_self_refine(self, data: pd.Series, user_instruction: str, dtypes: list[str], **kwargs):
        async with self.semaphore:
            self_refine_cost = 0.0
            generate_prompt = self.prompter.generate_prompt(data, user_instruction, self.output_columns, dtypes)
            output = await self.llm(generate_prompt, parse_tags=True, tags=self.output_columns, **kwargs)
            self_refine_cost += output["cost"]

            evaluate_prompt = self.prompter.generate_evaluate_prompt(data, output["raw_output"], user_instruction, dtypes)
            evaluate_output = await self.llm(evaluate_prompt, parse_tags=True, tags=["evaluation", "feedback"], **kwargs)
            self_refine_cost += evaluate_output["cost"]
            
            if "pass" in evaluate_output["evaluation"].lower():
                result = self._postprocess_map_output(output, self.output_columns)
                return result, self_refine_cost
            else:
                refine_prompt = self.prompter.generate_refine_prompt(data, output["raw_output"], user_instruction, self.output_columns, evaluate_output["feedback"], dtypes)
                refine_output = await self.llm(refine_prompt, parse_tags=True, tags=self.output_columns, **kwargs)
                self_refine_cost += refine_output["cost"]
                result = self._postprocess_map_output(refine_output, self.output_columns)
                refine_output["cost"] = self_refine_cost
                return result, self_refine_cost

    async def _execute_by_func(self, data: pd.Series, user_instruction: str, func: Callable, llm_call: Callable, **kwargs):
        try:
            map_result = func(data)
            result = self._postprocess_map_output(map_result, self.output_columns)
            return result, 0.0
        except Exception as e:
            return await llm_call(data, user_instruction)
        
    def _postprocess_map_output(self, llm_result: dict[str, str] | None, output_columns: list[str]) -> dict[str, Any] | None:
        try:
            if llm_result is None:
                return {column: None for column in output_columns}
            else:
                map_output = {column: llm_result.get(column, None) for column in output_columns}
                return map_output
        except KeyError as e:
            raise KeyError(f"Expected field {e} not found in the LLM response.")

    async def execute(
        self, 
        input_data: pd.DataFrame,
        **kwargs
    ):  
        if self.user_instruction is None and not self.has_udf():
            raise ValueError("Neither `user_instruction` nor `func` is given.")
        
        if input_data.empty:
            return MapOpOutputs(outputs={}, cost=0.0)

        processed_data = input_data[self.input_columns]
        dtypes = []
        for col in self.input_columns:
            if isinstance(input_data[col].dtype, ImageDtype):
                dtypes.append("image")
            else:
                dtypes.append("text")
        
        if self.strategy == "plain":
            execution_func = functools.partial(self._execute_by_plain_llm, dtypes=dtypes, model=self.model, **kwargs)
        elif self.strategy == "fewshot":
            assert self.context is not None, "Few-shot examples must be provided in the context for in-context learning."
            demos = self.context
            execution_func = functools.partial(self._execute_by_fewshot_llm, dtypes=dtypes, demos=demos, model=self.model, **kwargs)
        elif self.strategy == "self_refine":
            execution_func = functools.partial(self._execute_by_self_refine, dtypes=dtypes, model=self.model, **kwargs)
        else:
            raise ValueError(f"The optional strategies available for map are {self.strategy_options}. Strategy {self.strategy} is not supported.")

        # Create tasks for all data points
        tasks = []
        empty_map_result = {column: None for column in self.output_columns}
        for _, data in processed_data.iterrows():
            if data.empty:
                tasks.append(asyncio.create_task(asyncio.sleep(0, result=(empty_map_result, 0.0))))
            elif self.has_udf():
                tasks.append(asyncio.create_task(self._execute_by_func(data, self.user_instruction, self.tool, execution_func)))
            else:
                tasks.append(asyncio.create_task(execution_func(data, self.user_instruction)))
        
        # Wait for all tasks to complete
        if self.limit is not None and self.limit <= 0:
            warnings.warn("The limit should be positive. To execute, the limit will be ignored.")
            self.limit = None
        
        token_cost: float = 0.0
        map_outputs: dict[str, list] = {column: [] for column in self.output_columns}
        if self.limit is not None:
            num_passed_records: int = 0
            reach_limit: bool = False
            for i in range(0, len(tasks), self.limit):
                if reach_limit:
                    break
                batch_tasks = tasks[i:i + self.limit]
                batch_results = await asyncio.gather(*batch_tasks)
                token_cost += sum([result[1] for result in batch_results])
                for result, _ in batch_results:
                    num_passed_records += 1
                    for column in self.output_columns:
                        map_outputs[column].append(result[column])
                    if num_passed_records >= self.limit:
                        reach_limit = True
                        break
            num_remaining_records = len(processed_data) - num_passed_records
            if num_remaining_records > 0:
                for column in self.output_columns:
                    map_outputs[column].extend([None] * num_remaining_records)
        else:
            results = await asyncio.gather(*tasks)
            token_cost += sum([result[1] for result in results])
            for result, _ in results:
                for column in self.output_columns:
                    map_outputs[column].append(result[column])
        
        return MapOpOutputs(
            outputs=map_outputs,
            cost=token_cost
        )



import asyncio
from typing import Any, Iterable, Callable, Literal
from dataclasses import dataclass, field
import pandas as pd
from pandas.api.types import is_numeric_dtype

from nirvana.dataframe.arrays.image import ImageDtype
from nirvana.executors.tools import BaseTool, FunctionCallTool
from nirvana.ops.base import BaseOpOutputs, BaseOperation
from nirvana.ops.prompt_templates.reduce_prompter import ReducePrompter


def reduce_wrapper(
    input_data: Iterable[Any],
    user_instruction: str = None,
    input_column: str = None,
    context: list[dict] | str | None = None,
    model: str | None = None,
    func: Callable = None,
    strategy: Literal["plain"] = "plain",
    rate_limit: int = 16,
    assertions: list[Callable] | None = [],
    **kwargs
):
    """
    A function wrapper for reduce operation

    Args:
        input_data (Iterable[Any]): Input data
        user_instruction (str, optional): User instruction. Defaults to None.
        input_column (str, optional): Input column. Defaults to None.
        context (list[dict] | str, optional): Context. Defaults to None.
        model (str, optional): Model. Defaults to None.
        func (Callable, optional): User function. Defaults to None.
        strategy (Literal["plain"], optional): Strategy. Defaults to "plain".
        rate_limit (int, optional): Rate limit. Defaults to 16.
        assertions (list[Callable], optional): Assertions. Defaults to [].
        **kwargs: Additional keyword arguments for OpenAI Clent.
    """
    
    reduce_op = ReduceOperation(
        user_instruction=user_instruction,
        input_columns=[input_column],
        context=context,
        model=model,
        tool=FunctionCallTool.from_function(func=func) if func else None,
        strategy=strategy,
        rate_limit=rate_limit,
        assertions=assertions,
    )
    outputs = asyncio.run(reduce_op.execute(
        input_data=input_data,
        **kwargs
    ))
    return outputs


@dataclass
class ReduceOpOutputs(BaseOpOutputs):
    output: Any = field(default=None)

    def __add__(self, other: "ReduceOpOutputs"):
        return ReduceOpOutputs(
            output=self.output + other.output,
            cost=self.cost + other.cost
        )


class ReduceOperation(BaseOperation):
    """
    Reduce operator: Aggregates values in a column based on an NL-specified reduction function

    NOTE: This is a simple implementation that does not consider the case where the input length exceeds the token limit. 
          The next step is to implement several optimizations, like `summarize and aggregate` and `incremental aggregation`
    """

    def __init__(
        self,
        user_instruction: str = "",
        input_columns: list[str] = [],
        context: list[dict] | str | None = None,
        model: str | None = None,
        tool: Callable | BaseTool | None = None,
        strategy: Literal["plain"] = "plain",
        rate_limit: int = 16,
        assertions: list[Callable] | None = [],
    ):
        if tool and not isinstance(tool, BaseTool):
            tool = FunctionCallTool.from_function(func=tool)
        
        super().__init__(
            op_name="reduce",
            user_instruction=user_instruction,
            context=context,
            model=model,
            tool=tool,
            strategy=strategy,
            rate_limit=rate_limit,
            assertions=assertions,
        )
        self.prompter = ReducePrompter()
        self.input_columns = input_columns

    @property
    def dependencies(self) -> list[str]:
        return self.input_columns
    
    @property
    def generated_fields(self) -> list[str]:
        return []
    
    @property
    def op_kwargs(self):
        kwargs = super().op_kwargs
        kwargs["input_columns"] = self.input_columns
        return kwargs

    async def _execute_by_plain_llm(self, processed_data: Iterable[Any], user_instruction: str, dtype: str, **kwargs):
        async with self.semaphore:
            full_prompt = self.prompter.generate_prompt(processed_data, user_instruction, dtype)
            output = await self.llm(full_prompt, parse_tags=True, tags=["output"], **kwargs)
            return output["output"], output["cost"]
    
    async def _execute_by_func(self, processed_data: Iterable[Any], user_instruction: str, func: Callable, llm_call: Callable, dtype: str, **kwargs):
        try:
            reduce_results = processed_data.agg(func)
            return reduce_results, 0
        except:
            return await llm_call(processed_data, user_instruction, dtype, **kwargs)

    async def execute(
        self,
        input_data: pd.DataFrame,
        **kwargs
    ):
        if self.user_instruction is None and not self.has_udf():
            raise ValueError("Neither `user_instruction` nor `func` is given.")
        
        if input_data.empty:
            return ReduceOpOutputs(output="No data to process.")

        processed_data = input_data[self.input_columns[0]]
        if isinstance(processed_data.dtype, ImageDtype):
            dtype = "image"
        elif is_numeric_dtype(processed_data):
            dtype = "numeric"
        else:
            dtype = "text"

        reduce_results, token_cost = None, 0
        if self.has_udf() and dtype == "numeric":
            reduce_results, token_cost = await self._execute_by_func(processed_data, self.user_instruction, self.tool, self._execute_by_plain_llm, dtype, model=self.model, **kwargs)    
        else:
            reduce_results, token_cost = await self._execute_by_plain_llm(processed_data, self.user_instruction, dtype, model=self.model, **kwargs)

        return ReduceOpOutputs(
            output=reduce_results,
            cost=token_cost
        )



from nirvana.ops.prompt_templates.map_prompter import MapPrompter
from nirvana.ops.prompt_templates.filter_prompter import FilterPrompter
from nirvana.ops.prompt_templates.reduce_prompter import ReducePrompter
from nirvana.ops.prompt_templates.join_prompter import JoinPrompter

__all__ = [
    "MapPrompter",
    "FilterPrompter",
    "ReducePrompter",
    "JoinPrompter",
]



import os
import base64
import requests
from typing import Sequence
import numpy as np
from pandas.api.extensions import ExtensionDtype, ExtensionArray


def load_audio(raw_data: str | None) -> bytes | None:
    if raw_data is None:
        return None
    
    if raw_data.startswith("https://"):
        response = requests.get(raw_data)
        response.raise_for_status()
        audio_obj = response.content
        return base64.b64encode(audio_obj).decode('utf-8')
    elif os.path.isfile(raw_data):
        with open(raw_data, "rb") as f:
            file_obj = f.read()
            return base64.b64encode(file_obj).decode("utf-8")
    else:
        raise ValueError(f"Unrecognized file input, support local path, http url, and S3, got {raw_data}")


class AudioDtype(ExtensionDtype):
    """
    Dtype for AudioArray.
    """
    name = "audio"
    type = str
    na_value = None

    def __repr__(self):
        return "dtype('audio')"

    @classmethod
    def construct_array_type(cls):
        return AudioArray


class AudioArray(ExtensionArray):
    """
    An array where each element is a file path (a string).
    """
    def __init__(self, data):
        self._data = np.asarray(data, dtype=str)
        self._dtype = AudioDtype()

    def __getitem__(self, item: int | slice | Sequence[int]):
        if isinstance(item, int):
            return self._data[item]
        else:
            return AudioArray(self._data[item])

    def __len__(self):
        return len(self._data)

    @property
    def dtype(self) -> AudioDtype:
        return self._dtype

    def isna(self):
        return np.array([x is None for x in self._data], dtype=bool)

    def take(self, indices: Sequence[int], allow_fill=False, fill_value=None):
        data = self._data.take(self._data, indices, allow_fill=allow_fill, fill_value=fill_value)
        return AudioArray(data)

    def copy(self):
        return AudioArray(self._data.copy())

    def _concat_same_type(cls, to_concat):
        data = np.concatenate([arr._data for arr in to_concat])
        return cls(data)



import logging
import asyncio
from functools import partial
import time
import numpy as np
import pandas as pd

from nirvana.executors.llm_backbone import LLMClient
from nirvana.lineage.abstractions import LineageNode, execute_node
from nirvana.optim.rules import (
    FilterPushdown,
    FilterPullup,
    MapPullup,
    NonLLMPushdown,
    NonLLMReplace,
    OperatorFusion,
)
from nirvana.optim.evaluator import Evaluator

logger = logging.getLogger(__name__)


class PlanCost:
    error_tolerance = 0.8

    def __init__(self, plan: LineageNode, rules: set, accuracy: float, cost: float, runtime: float):
        self.plan = plan
        self.rules = rules
        self.accuracy = accuracy
        self.cost = cost
        self.runtime = runtime

    def __lt__(self, other: "PlanCost") -> bool:
        if self.accuracy / (other.accuracy + 1e-6) >= self.error_tolerance:
            return self.cost < other.cost
        return False
    
    def __le__(self, other: "PlanCost") -> bool:
        if self.accuracy / (other.accuracy + 1e-6) >= self.error_tolerance:
            return self.cost <= other.cost
        return False
    
    def __gt__(self, other: "PlanCost") -> bool:
        if self.accuracy / (other.accuracy + 1e-6) >= self.error_tolerance:
            return self.cost > other.cost
        return True
    
    def __ge__(self, other: "PlanCost") -> bool:
        if self.accuracy / (other.accuracy + 1e-6) >= self.error_tolerance:
            return self.cost >= other.cost
        return True
    
    def __eq__(self, other: "PlanCost") -> bool:
        if self.accuracy / other.accuracy >= self.error_tolerance:
            return self.cost == other.cost
        return False


class LogicalOptimizer:

    def __init__(
        self,
        max_round: int = 5,
        agent: LLMClient = None,
        filter_pullup: bool = True,
        filter_pushdown: bool = True,
        map_pullup: bool = True,
        non_llm_pushdown: bool = True,
        non_llm_replace: bool = True,
        operator_fusion: bool = True,
    ):
        self.agent = agent
        self.max_round = max_round
        if non_llm_replace or operator_fusion:
            assert agent is not None, "LLM must be provided for semantic-aware plan rewrites, e.g., for Non-LLM Replace and Operator Fusion."
        self.rules = set()
        if filter_pullup:
            self.rules.add(FilterPullup.transform)
        if filter_pushdown:
            self.rules.add(FilterPushdown.transform)
        if map_pullup:
            self.rules.add(MapPullup.transform)
        if non_llm_pushdown:
            self.rules.add(NonLLMPushdown.transform)
        if non_llm_replace:
            self.rules.add(partial(NonLLMReplace.transform, rewriter=self.agent))
        if operator_fusion:
            self.rules.add(partial(OperatorFusion.transform, rewriter=self.agent))
        self.plan_candidates = []

    def clear(self):
        self.rules = set()
        self.plan_candidates = []

    def sample_from_candidates(self, lambda_=0.2):
        token_costs = np.array([plan.cost for plan in self.plan_candidates])
        token_costs = (np.max(token_costs) - token_costs) / (np.max(token_costs) - np.min(token_costs) + 1e-6)
        costs_prob = np.exp(token_costs) / np.exp(token_costs).sum()
        uniform_prob = np.ones_like(token_costs, dtype=float) / len(token_costs)
        total_prob = lambda_ * uniform_prob + (1 - lambda_) * costs_prob

        selected_index = np.random.choice(len(token_costs), p=total_prob)
        return self.plan_candidates[selected_index]

    def preprare_for_logical_optimization(
        self,
        plan: LineageNode,
        dev_datasets: list | None,
        num_samples: int | None,
    ):
        initial_state = []
        def prepare_dev_datasets(node: LineageNode):
            if node.left_child:
                prepare_dev_datasets(node.left_child)
            if node.right_child:
                prepare_dev_datasets(node.right_child)
            
            if node.op_name == "scan":
                initial_state.append(node.datasource)
                if dev_datasets:
                    node.datasource = dev_datasets.pop(0)
                elif num_samples is not None:
                    node.operator.set_sample_size(num_samples)
                else:
                    raise ValueError(
                        "Either `dev_datasets` or `num_samples` must be provided to prepare the development datasets for logical optimization."
                    )
                return
        prepare_dev_datasets(plan)
        return initial_state

    def recover_plan_state(self, initial_state: list, plan: LineageNode):
        def reset_datasources(node: LineageNode):
            if node.left_child:
                reset_datasources(node.left_child)
            if node.right_child:
                reset_datasources(node.right_child)
            
            if node.op_name == "scan":
                node.datasource = initial_state.pop(0)
                node.operator.set_sample_size(None)
                return
        reset_datasources(plan)
    
    async def estimate_plan_cost(self, plan: LineageNode, ground_truth: pd.DataFrame = None):
        try:
            results, token_cost, exec_time = await execute_node(plan)
        except Exception as e:
            return None, 0.0, 0.0, 0.0
        if ground_truth is None:
            accuracy_score = 1.0
        else:
            accuracy_score = await Evaluator.evaluate(ground_truth, results, self.agent)
        return results, accuracy_score, token_cost, exec_time
    
    async def _optimize(self, initial_plan: LineageNode, dev_datasets: list | None = None, num_samples: int = 10):
        round = 0
        optimize_cost = 0.0
        optimize_start_time = time.time()
        # 0. prepare for logical optimization
        initial_state = self.preprare_for_logical_optimization(
            initial_plan,
            dev_datasets=dev_datasets,
            num_samples=num_samples,
        )
        
        # 1. get the data processing ground truth by executing the initial plan on the validation set
        groundtruth, accuracy_score, token_cost, exec_time = await self.estimate_plan_cost(initial_plan)
        init_plan_cost = PlanCost(initial_plan, set(), accuracy_score, token_cost, exec_time)
        self.plan_candidates.append(init_plan_cost)

        # 2. optimize the plan
        while round < self.max_round:
            logger.info(f"Round {round + 1} of logical optimization is started.")
            # 2.1. sample a plan from the candidate list
            cand_plan_cost = self.sample_from_candidates()

            # 2.2. select a transformation rule that has not been applied to the sampled plan
            applicable_rules = self.rules - cand_plan_cost.rules
            if len(applicable_rules) == 0:
                round += 1
                continue

            selected_rule = np.random.choice(list(applicable_rules))
            optimized_plan, cost_per_rewrite = await selected_rule(cand_plan_cost.plan)
            optimize_cost += cost_per_rewrite

            # 2.3. compare the results with the ground truth
            _, accuracy_score, token_cost, exec_time = await self.estimate_plan_cost(optimized_plan, groundtruth)
            applied_rules = cand_plan_cost.rules.union({selected_rule})
            plan_cost = PlanCost(optimized_plan, applied_rules, accuracy_score, token_cost, exec_time)
            self.plan_candidates.append(plan_cost)
            round += 1

        # 3. select the best plan from the candidate list
        best_plan = sorted(self.plan_candidates)[0]
        optimize_end_time = time.time()
        optimize_time = optimize_end_time - optimize_start_time

        # 4. restore the state for the final best plan
        self.recover_plan_state(initial_state, best_plan.plan)

        logger.info(f"Plan optimization is finished, taking {optimize_time:.4f} seconds and ${optimize_cost:.4f}. Here are some statistics:")
        logger.info(f"initial plan cost: {init_plan_cost.cost} -> optimized plan cost: {best_plan.cost}")
        logger.info(f"initial plan runtime: {init_plan_cost.runtime} -> optimized plan runtime: {best_plan.runtime}")
        logger.info(f"initial plan accuracy: {init_plan_cost.accuracy} -> optimized plan accuracy: {best_plan.accuracy}")        
        return best_plan.plan

    def optimize(
        self,
        plan: LineageNode,
        dev_datasets: list | None = None,
        num_samples: int = 10,
    ):
        return asyncio.run(self._optimize(plan, dev_datasets, num_samples))



import sys
import base64
import requests # type: ignore
from io import BytesIO
from typing import Union, Sequence
import numpy as np
import pandas as pd
from PIL import Image
from pandas.api.extensions import ExtensionDtype, ExtensionArray


def load_image(raw_data: str | bytes | Image.Image | None) -> str | None:
    if raw_data is None:
        return None

    if isinstance(raw_data, Image.Image):
        buffered = BytesIO()
        raw_data.save(buffered, format="PNG")
        return "data:image/png;base64," + base64.b64encode(buffered.getvalue()).decode("utf-8")
    elif isinstance(raw_data, bytes):
        return "data:image/png;base64," + base64.b64encode(raw_data).decode("utf-8")
    elif isinstance(raw_data, str):
        if raw_data.startswith("data:image"):
            return raw_data
        elif raw_data.startswith("https://"):
            return raw_data
        elif raw_data.startswith("s3://"):
            from botocore.exceptions import NoCredentialsError, PartialCredentialsError

            try:
                import boto3

                s3 = boto3.client("s3")
                bucket_name, key = raw_data[5:].split("/", 1)  # Split after "s3://"
                response = s3.get_object(Bucket=bucket_name, Key=key)
                image_data = response["Body"].read()
                # is image_data bytes?
                image_obj = Image.open(BytesIO(image_data))
                buffered = BytesIO()
                image_obj.save(buffered, format="PNG")
                return "data:image/png;base64," + base64.b64encode(buffered.getvalue()).decode("utf-8")
            except (NoCredentialsError, PartialCredentialsError) as e:
                raise ValueError("AWS credentials not found or incomplete.") from e
            except Exception as e:
                raise ValueError(f"Failed to fetch image from S3: {e}") from e
        else:
            # it's a local file path
            with open(raw_data, "rb") as f:
                image_data = f.read()
                return "data:image/png;base64," + base64.b64encode(image_data).decode("utf-8")
    else:
        raise ValueError(f"Unrecognized image input, support local path, http url, base64, S3, and PIL.Image, got {raw_data}")


class ImageDtype(ExtensionDtype):
    name = 'image'
    type = str
    na_value = None

    def __repr__(self):
        return "dtype('image')"

    @classmethod
    def construct_array_type(cls):
        return ImageArray


class ImageArray(ExtensionArray):
    def __init__(self, values):
        self._data = np.asarray(values, dtype=object)
        self._dtype = ImageDtype()

    def __getitem__(self, item: Union[int, slice, Sequence[int]]) -> np.ndarray:
        result = self._data[item]

        if isinstance(item, (int, np.integer)):
            # Return the raw value for display purposes
            return result

        return ImageArray(result)

    def __setitem__(self, key, value) -> None:
        """Set one or more values inplace, with cache invalidation."""
        if isinstance(key, np.ndarray):
            if key.dtype == bool:
                key = np.where(key)[0]
            key = key.tolist()
        if isinstance(key, (int, np.integer)):
            key = [key]
        if isinstance(key, slice):
            key = range(*key.indices(len(self)))
        if hasattr(value, "__iter__") and not isinstance(value, (str, bytes)):
            for idx, val in zip(key, value):
                self._data[idx] = load_image(val)
        else:
            for idx in key:
                self._data[idx] = load_image(value)

    def get_image(self, index: int) -> str | None:
        return self._data[index]

    def isna(self) -> np.ndarray:
        return pd.isna(self._data)

    def take(self, indices: Sequence[int], allow_fill: bool = False, fill_value=None) -> "ImageArray":
        result = self._data.take(indices, axis=0)
        if allow_fill and fill_value is not None:
            result[indices == -1] = fill_value
        return ImageArray(result)

    def copy(self) -> "ImageArray":
        new_array = ImageArray(self._data.copy())
        return new_array

    def _concat_same_type(cls, to_concat: Sequence["ImageArray"]) -> "ImageArray":
        """
        Concatenate multiple ImageArray instances into a single one.

        Args:
            to_concat (Sequence[ImageArray]): A sequence of ImageArray instances to concatenate.

        Returns:
            ImageArray: A new ImageArray containing all elements from the input arrays.
        """
        # create list of all data
        combined_data = np.concatenate([arr._data for arr in to_concat])
        return cls(combined_data)
    
    @classmethod
    def _from_sequence(cls, scalars, dtype=None, copy=False):
        if copy:
            scalars = np.array(scalars, dtype=object, copy=True)
        return cls(scalars)

    def __len__(self) -> int:
        return len(self._data)

    def __eq__(self, other) -> np.ndarray:  # type: ignore
        if hasattr(other, "__iter__") and not isinstance(other, str):
            if len(other) != len(self):
                return np.repeat(False, len(self))
            return np.array([img1 == img2 for img1, img2 in zip(self._data, other)], dtype=bool)
        return np.array([img == other for img in self._data], dtype=bool)

    @property
    def dtype(self) -> ImageDtype:
        return self._dtype

    @property
    def nbytes(self) -> int:
        return sum(sys.getsizeof(val) for val in self._data if val)

    def __repr__(self) -> str:
        return f"ImageArray([{', '.join([f'<Image: {img}>' if img is not None else 'None' for img in self._data[:5]])}, ...])"

    def _formatter(self, boxed: bool = False):
        return lambda x: f"<Image: {x}>" if x is not None else "None"

    def to_numpy(self, dtype=None, copy=False, na_value=None) -> np.ndarray:
        """Convert the ImageArray to a numpy array."""
        return self._data

    def __array__(self, dtype=None) -> np.ndarray:
        """Numpy array interface."""
        return self.to_numpy(dtype=dtype)



FILTER_PUSHDOWN = """- Filter pushdown: Move a filter operator that does not rely on results of prior operator to occur before the prior operator."""


OPERATOR_FUSION = """- Operator fusion: Merge multiple operators that are in the same category and work on the same column into one operator. To keep semantically equivalent, you are required to rewrite the instruction for the new operator."""


NON_LLM_REPLACEMENT = """- Non-LLM operator replacement: Replace the NL instruction over non-image/video/audio data with an equivalent compute function only when `user_instruction` can be converted to a straightforward `func` (`func` has to be a simple lambda or built-in function). Note that the `user_instruction` should be preserved in the new operator."""


PLAN_OPIMIZE_PROMPT = """Given a dataset and a user-specified logical plan for a data processing task, you are required to optimize the plan by applying a set of equivalent transformations.
The transformations should be semantically equivalent to the original plan but can improve efficiency and reduce runtime costs/LLM calls.
The dataset is given in the form of a pandas dataframe, and a logical plan is represented by a sequence of .semantic_*(...) functions (* can be replaced with names of possible operators).
Here is an example that defines a logical plan containing map and filter operations:
```python
# df is the name of the dataset
df.semantic_map(user_instruction="map instruction", input_column="col_a", output_column="col_b")
df.semantic_filter(user_instruction="filter instruction", input_column="col_c")
```

The supported operators and their required arguments are as follows.
1. map: Perform an element-wise projection specified by natural language on a given column to a new column. Required arguments:
- user_instruction: a natural language expression
- func: Python function, returns a single value from a single value
- input_column: the name of the column on which the operation is performed
- output_column: the name of the new column that the operation generates
2. filter: Evaluate a natural language condition per value in a given column (returning boolean). Required arguments:
- user_instruction: the natural language condition
- func: Python function, returns boolean value by a predicate
- input_column: the name of column on which the operation is performed
3. join: Join a table with another table by keeping all tuple pairs that satisfy a natural language condition. Required arguments:
- other: the other dataset to join with
- user_instruction: the join condition in natural language
- left_on: the name of the column from the left table to join on
- right_on: the name of the column from the right table to join on
- how: the type of join to be performed (e.g., inner, left, right)
- func: function to evaluate the join condition
4. reduce: Aggregate multiple values in a given column into a single result. Required arguments:
- user_instruction: the reducer function in natural language
- func: function to use for aggregating the data
- input_column: the name of column on which the operation is performed

The available transformation rules are as follows.
- Operator fusion: Fuse multiple operators that are in the same category and work on the same column into one operator. To keep semantically equivalent, you are required to rewrite the instruction for the new operator.
- Non-LLM operator replacement: Replace the NL instruction over non-image/video/audio data with an equivalent compute function only when `user_instruction` can be converted to a straightforward `func` (`func` has to be a simple lambda or built-in function). Note that the `user_instruction` should be preserved in the new operator.

Now, you are given two datasets with columns: {columns}, and a logical plan:
```python
{logical_plan}
```

Optimize the logical plan by applying as many transformation rules as possible but transformations rigidly conform to the rules. Note that **any modification to the pre-defined operator interfaces is not allowed**, and **the argument input_column for the current operation can only be selected from the dataset's columns or output_column of previous operations.**
The optimized plan is output as executable python code. Note that **each line in the code block represents a single complete function call.** If no further optimization proposed, return an empty code block. ONLY ONE code block can be contained in the output.
"""


RESULT_EVALUATE_PROMPT = """Here are a golden analysis result obtained by a golden data processing plan and a result derived from an alternative data processing plan.
Evaluate the two data analysis results and return a rating between 0 and 10, where 0 means the two results are completely different and 10 means they are exactly the same.

Ground truth:
{ground_truth}

Result from the alternative plan:
{result}

You should carefully consider all values (and their semantics) in both analysis results. The rating score is enclosed within <score></score> tags, i.e., <score>Rating Score</score>.
"""



from typing import Any, Callable, Awaitable
from abc import ABC, abstractmethod


class BaseTool(ABC):
    name: str
    description: str

    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)

    @abstractmethod
    def __call__(self, *args, **kwds) -> Any:
        pass


class PythonInterpreterTool(BaseTool):
    descrption = "Python Interpreter that runs code in a sandboxed environment"

    def __call__(self, *args, **kwds):
        pass


class FunctionCallTool(BaseTool):
    description = "This is a tool that evaluates python code. It can be used to perform calculations."

    def __init__(
            self,
            name: str | None = None,
            func: Callable | None = None,
            coroutine: Callable[..., Awaitable[Any]] | None = None,
            **kwargs
    ):
        self.name = name
        self.func = func
        self.coroutine = coroutine

    def __repr__(self) -> str:
        return f"FunctionCall(name={self.name})"
    
    def __str__(self) -> str:
        return self.__repr__()

    @classmethod
    def from_function(
        cls,
        func: Callable | None = None,
        coroutine: Callable[..., Awaitable[Any]] | None = None,
        name: str | None = None,
    ):
        if func is not None:
            func = func
        elif coroutine is not None:
            func = coroutine
        else:
            raise ValueError("Either `func` or `coroutine` must be provided.")
        
        name = name or func.__name__

        return cls(
            name=name,
            func=func,
            coroutine=coroutine
        )

    def _run(self, *args, **kwargs):
        return self.func(*args, **kwargs)
    
    async def _arun(self, *args, **kwargs):
        return await self.coroutine(*args, **kwargs)

    def __call__(self, *args, **kwargs):
        if self.func is not None:
            return self._run(*args, **kwargs)
        elif self.coroutine is not None:
            return self._arun(*args, **kwargs)



from typing import Any
import pandas as pd


class FilterPrompter:
    def __init__(self):
        self.system_instruction = (
            "You are a helpful assistant helping the user make sense of their data. "
            "You are performing a filter operation (data -> True or False) to "
            "determine whether the given data satisfies the given condition(s)."
        )
        self.output_format = (
            "The answer should strictly be either True or False. "
            "Output the result of the filter operation concisely in the following format.\n"
            "<output> True or False </output>\n"
        )

    def generate_prompt(
        self,
        data: pd.Series,
        user_instruction: str,
        dtypes: list[str]
    ):
        # 1. Prepare system message
        sys_message = [
            {"role": "system", "content": self.system_instruction},
            {"role": "system", "content": self.output_format}
        ]

        # 2. Prepare data
        user_content = []
        for dtype, (key, val) in zip(dtypes, data.items()):
            if dtype == "text":
                user_content.append({"type": "input_text", "text": f"{key}: {val}"})
            elif dtype == "image":
                user_content.append({"type": "input_text", "text": f"{key}:"})
                user_content.append({"type": "input_image", "image_url": val})
            elif dtype == "audio":
                user_content.append({"type": "input_text", "text": f"{key}:"})
                user_content.append(
                    {"type": "input_audio", "input_audio": {"data": val, "format": "wav"}}
                )
            else:
                raise ValueError(f"Data type {dtype} is not supported.")
        
        # 3. Prepare the given condition
        conditions = f"condition: {user_instruction}"
        user_content.append({"type": "input_text", "text": conditions})
        
        user_message = [{"role": "user", "content": user_content}]

        messages = sys_message + user_message
        return messages
    
    def generate_fewshot_prompt(
        self,
        data: pd.Series,
        user_instruction: str,
        dtypes: list[str],
        demos: list[dict[str, Any]]
    ):
        # 1. Prepare system message
        sys_message = [
            {"role": "system", "content": self.system_instruction},
            {"role": "system", "content": self.output_format}
        ]

        # 2. Prepare demonstration message
        demos_message = [{"type": "input_text", "text": f"Here are some examples:"}]
        for demo in demos:
            demo_content = []
            demo_data: pd.Series | dict = demo["data"]
            demo_answer: str = demo["answer"]
            for dtype, (key, val) in zip(dtypes, demo_data.items()):
                if dtype == "text":
                    demo_content.append({"type": "input_text", "text": f"{key}: {val}"})
                elif dtype == "image":
                    demo_content.append({"type": "input_text", "text": f"{key}:"})
                    demo_content.append({"type": "input_image", "image_url": val})
                elif dtype == "audio":
                    demo_content.append({"type": "input_text", "text": f"{key}:"})
                    demo_content.append(
                        {"type": "input_audio", "input_audio": {"data": val, "format": "wav"}}
                    )
                else:
                    raise ValueError(f"Data type {dtype} is not supported.")
            demo_content.append({"type": "input_text", "text": f"condition: {user_instruction}"})
            demo_content.append({"type": "input_text", "text": f"Answer: {demo_answer}"})
            demos_message.append(
                {"role": "assistant", "content": demo_content}
            )

        # 3. Prepare user message
        user_content = []
        for dtype, (key, val) in zip(dtypes, data.items()):
            if dtype == "text":
                user_content.append({"type": "input_text", "text": f"{key}: {val}"})
            elif dtype == "image":
                user_content.append({"type": "input_text", "text": f"{key}:"})
                user_content.append({"type": "input_image", "image_url": val})
            elif dtype == "audio":
                user_content.append({"type": "input_text", "text": f"{key}:"})
                user_content.append(
                    {"type": "input_audio", "input_audio": {"data": val, "format": "wav"}}
                )
            else:
                raise ValueError(f"Data type {dtype} is not supported.")
        
        # 4. Prepare the given condition
        conditions = f"condition: {user_instruction}"
        user_content.append({"type": "input_text", "text": conditions})
        user_message = [{"role": "user", "content": user_content}]
        
        messages = sys_message + demos_message + user_message
        return messages
    
    def generate_evaluate_prompt(
        self,
        data: pd.Series,
        answer: Any,
        user_instruction: str,
        dtypes: list[str]
    ):
        evaluator_task = [{"role": "user", "content": "Analyze the filter operation tasked with evaluating the condition on the given data:"}]

        _input = [{"type": "input_text", "text": "Input:"}]
        for dtype, (key, val) in zip(dtypes, data.items()):
            if dtype == "text":
                _input.append({"type": "input_text", "text": f"{key}: {val}"})
            elif dtype == "image":
                _input.append({"type": "input_text", "text": f"{key}:"})
                _input.append({"type": "input_image", "image_url": val})
            elif dtype == "audio":
                _input.append({"type": "input_text", "text": f"{key}:"})
                _input.append(
                    {"type": "input_audio", "input_audio": {"data": val, "format": "wav"}}
                )
            else:
                raise ValueError(f"Data type {dtype} is not supported.")
        
        _output = [{"type": "input_text", "text": f"Output:\n{answer}"}]

        instruction = user_instruction.strip() + " " + self.output_format
        _instruction = [{"type": "input_text", "text": f"Instruction:\n{instruction}"}]
        
        evaluator_context = [
            {"role": "user", "content": _input + _output + _instruction}
        ]

        evaluator_criteria = (
            "Evaluate the filter operation based on the following criteria:\n"
            "1. Is the final determination either True or False?\n"
            "2. Does the output strictly adhere to the required output format?\n"
            "3. Whether correctly derives the final True/False determination based on the given instruction.\n"
            "You should be evaluating only and not attemping to solve the task. Only output PASS if all criteria are met and you have no further suggestions for improvements.\n"
            "Output your evaluation concisely in the following format.\n"
            "<evaluation> PASS or FAIL </evaluation>\n"
            "<feedback> What needs improvement and why. </feedback>\n"
        )
        evaluator_criteria = [{"role": "user", "content": evaluator_criteria}]
        messages = evaluator_task + evaluator_context + evaluator_criteria
        return messages
    
    def generate_refine_prompt(
        self,
        data: pd.Series,
        answer: Any,
        user_instruction: str,
        feedback: str,
        dtypes: list[str]
    ):
        initial_generate_prompt = self.generate_prompt(data, user_instruction, dtypes)
        refine_context =(
            "There is feedback from your previous output.\n"
            f"Output:\n{answer}\nFeedback:\n{feedback}\n"
            "Your task is to refine the output based on the feedback."
        )
        refine_prompt = [{"role": "user", "content": refine_context}]
        messages = initial_generate_prompt + refine_prompt
        return messages



from nirvana.lineage.abstractions import LineageNode


def rewire_nodes(start_node: LineageNode, node_list: list[LineageNode]):
    last_node = start_node
    for node in node_list:
        node.node_fields.left_input_fields = last_node.node_fields.output_fields
        node.node_fields.output_fields = node.node_fields.left_input_fields
        node.set_left_child(last_node)
        last_node = node
    return last_node


class FilterPullup:
    @classmethod
    def pull_up(cls, node: LineageNode) -> tuple[LineageNode, list[LineageNode]]:
        if node.op_name == "scan":
            return node, []

        elif node.op_name == "filter":
            child_node, pullup_filters = cls.pull_up(node.left_child)
            pullup_filters.append(node)
            last_node = rewire_nodes(child_node, pullup_filters)
            return child_node, pullup_filters

        elif node.op_name == "join":
            left_child_node, left_pullup_filters = cls.pull_up(node.left_child)
            right_child_node, right_pullup_filters = cls.pull_up(node.right_child)

            node.set_left_child(left_child_node)
            node.set_right_child(right_child_node)
            pullup_filters = left_pullup_filters + right_pullup_filters
            last_node = rewire_nodes(node, pullup_filters)
            return node, pullup_filters

        elif node.op_name in ["map", "rank"]:
            child_node, pullup_filters = cls.pull_up(node.left_child)
            # Update operator's left input fields to what the child node previously produced
            node.node_fields.left_input_fields = child_node.node_fields.output_fields
            node.node_fields.output_fields = list(
                set(node.node_fields.left_input_fields + node.operator.generated_fields)
            )
            node.set_left_child(child_node)
            last_node = rewire_nodes(node, pullup_filters)
            return node, pullup_filters

        elif node.op_name == "reduce":
            child_node, pullup_filters = cls.pull_up(node.left_child)
            last_node = rewire_nodes(child_node, pullup_filters)
            node.node_fields.left_input_fields = last_node.node_fields.output_fields
            node.set_left_child(last_node)
            return node, pullup_filters
    
    @classmethod
    async def transform(cls, node: LineageNode) -> tuple[LineageNode, float]:
        node, _ = cls.pull_up(node)
        return node, 0.0



"""
Record the OP lineage (operator and its user instruction) for optimizing operator orchestration.
"""
import copy
import time
import asyncio
from collections import deque

from nirvana.lineage.abstractions import LineageNode, execute_node, collect_op_metadata
from nirvana.optim.optimizer import PlanOptimizer, OptimizeConfig

class LineageMixin:
    def initialize(self):
        op_kwargs = {"source": "dataframe", "output_columns": self.columns}
        data_kwargs = {"left_input_fields": [], "right_input_fields": [], "output_fields": self.columns}
        node = LineageNode(op_name="scan", op_kwargs=op_kwargs, node_fields=data_kwargs, datasource=self._data)
        self.leaf_node = node

    def add_operator(self, op_name: str, op_kwargs: dict, data_kwargs: dict, **kwargs):
        node = LineageNode(op_name, op_kwargs=op_kwargs, node_fields=data_kwargs)
        if op_name == "join":
            node.set_left_child(self.leaf_node)
            node.set_right_child(kwargs["other"].leaf_node)
            self.leaf_node = node
        else:
            node.set_left_child(self.leaf_node)
            self.leaf_node = node

    def create_plan_optimizer(self, config: OptimizeConfig = None):
        self.optimizer = PlanOptimizer(config)

    def execute(self):
        execution_start_time = time.time()
        dataframe_from_node, token_cost = asyncio.run(execute_node(self.leaf_node))
        execution_end_time = time.time()
        execution_time = execution_end_time - execution_start_time
        return dataframe_from_node, token_cost, execution_time
    
    def print_lineage_graph(self, op_signature_width: int = 512, max_instruction_print_length: int = 256):
        lineage_graph_strings = []
        op_strings_in_same_hop = []
        node_queue = deque([self.leaf_node])

        while node_queue:
            node = node_queue.popleft()
            if node is None:
                lineage_graph_strings.append(op_strings_in_same_hop)
                op_strings_in_same_hop = []
                continue

            op_info = collect_op_metadata(node, max_instruction_print_length)
            op_strings_in_same_hop.append(op_info)

            node_queue.append(None)
            node_queue.append(node.left_child)
            if node.right_child:
                node_queue.append(node.right_child)

        stringified_lineage_graph = ""
        while lineage_graph_strings:
            ops_info = lineage_graph_strings.pop()
            ops_info_string = ""
            for op_info in ops_info:
                op_info = f"{op_info:<{op_signature_width}}"
                ops_info_string += op_info
            divider = f"{'|':<{op_signature_width}}\t" * len(ops_info)
            stringified_lineage_graph += ops_info_string.strip() + "\n"
            stringified_lineage_graph += divider + "\n"

        print(f"Lineage Graph:\n{stringified_lineage_graph}")

    def clear_lineage_graph(self):
        def _delete_node(node: LineageNode):
            if node.left_child:
                _delete_node(node.left_child)
            if node.right_child:
                _delete_node(node.right_child)
            del node
            return
        
        if self.optimizer:
            self.optimizer.clear()
        if self.leaf_node:
            temp_node = copy.copy(self.leaf_node)
            self.leaf_node = None
            # See join left and right tables in data lineage, 
            # empty_lineage will delete all nodes along two upstream sub-lineages
            # So put a note here if there is a bug when deleting nodes
            _delete_node(temp_node)
            del temp_node



import warnings
from typing import Optional
from pydantic import BaseModel, Field
from typing_extensions import deprecated
import pandas as pd

from nirvana.executors.llm_backbone import LLMClient
from nirvana.lineage.abstractions import LineageNode
from nirvana.optim.logical import LogicalOptimizer
from nirvana.optim.physical import PhysicalOptimizer
    

class OptimizeConfig(BaseModel):
    do_logical_optimization: bool = Field(default=True, description="whether perform logical plan optimization.")
    do_physical_optimization: bool = Field(default=True, description="whether perform physical plan optimization.")
    sample_size: Optional[int] = Field(default=5, description="The number of data used for plan optimization.")
    max_rounds: int = Field(default=5, description="The maximum number of optimization rounds for logical optimization.")
    improve_margin: float = Field(default=0.2, description="The improvement margin for physical optimization.")

    # transformation rules
    filter_pullup: bool = Field(default=True, description="Whether use filter pullup.")
    filter_pushdown: bool = Field(default=True, description="Whether use filter pushdown.")
    map_pullup: bool = Field(default=True, description="Whether use map pullup.")
    non_llm_pushdown: bool = Field(default=True, description="Whether use non-llm pushdown.")
    non_llm_replace: bool = Field(default=True, description="Whether use non-llm replacement")
    operator_fusion: bool = Field(default=True, description="Whether use operator fusion.")

    # available backend models for query optimization
    avaiable_models: list[str] = Field(default_factory=list, description="The available models for physical optimization.")

    # deprecated fields
    approx_mode: bool = Field(
        default=True,
        description="Whether use approximation for physical optimization.",
        deprecated=deprecated("`approx_mode` is deprecated and will be removed in future versions."),
    )
    sample_ratio: Optional[float] = Field(
        default=None,
        description="The ratio of data used for physical optimization.",
        deprecated=deprecated("`sample_ratio` is deprecated and will be removed in future versions. Please use `sample_size` instead."),
    )


class PlanOptimizer:
    client: LLMClient = None

    @classmethod
    def set_agent(cls, client: LLMClient):
        cls.client = client

    def __init__(self, config: OptimizeConfig | dict = None):
        if isinstance(config, dict):
            config = OptimizeConfig(**config)
        self.config = config if config is not None else OptimizeConfig()
        self.prepare_optimizers(self.config)

    def set_config(self, config: OptimizeConfig):
        self.config = config

    def prepare_optimizers(self, config: OptimizeConfig):
        if self.config.sample_ratio:
            warnings.warn(
                "`sample_ratio` is deprecated and will be removed in future versions. Now we use default `sample_size` instead.",
                DeprecationWarning,
            )
        if config.do_logical_optimization:
            self.logical_optimizer = LogicalOptimizer(
                self.config.max_rounds, self.client, config.filter_pullup, config.filter_pushdown, config.map_pullup, config.non_llm_pushdown, config.non_llm_replace
            )
        else:
            self.logical_optimizer = None

        if config.do_physical_optimization:
            self.physical_optimizer = PhysicalOptimizer(self.client, config.avaiable_models, config.sample_size, config.improve_margin)
        else:
            self.physical_optimizer = None

    def clear(self):
        if self.logical_optimizer:
            self.logical_optimizer.clear()

    def optimize_logical_plan(self, plan: LineageNode):
        plan = self.logical_optimizer.optimize(plan, num_samples=self.config.sample_size)
        return plan
    
    def optimize_physical_plan(self, plan: LineageNode):
        return self.physical_optimizer.optimize(plan)



import logging
from typing import Union
import pandas as pd

from nirvana.dataframe.frame import DataFrame
from nirvana.dataframe.arrays.image import ImageArray, ImageDtype
from nirvana.dataframe.arrays.audio import AudioArray, AudioDtype
from nirvana.dataframe.arrays.file import FileArray, FileDtype
from nirvana.utils import configure_llm_backbone

pd.api.extensions.register_extension_dtype(ImageDtype)
pd.api.extensions.register_extension_dtype(AudioDtype)
pd.api.extensions.register_extension_dtype(FileDtype)


logging.basicConfig(format="[\033[34m%(asctime)s\033[0m] %(message)s", datefmt="%Y-%m-%d %H:%M:%S", level=logging.INFO)
logger = logging.getLogger(__name__)


def convert_to_base_data(data: Union[pd.Series, list]) -> list:
    """
    Converts data to proper base data type.
    - For original pandas data types, this is returns tolist().
    - For ImageDtype, this returns list of PIL.Image.Image.
    """
    if isinstance(data, pd.Series):
        if isinstance(data.dtype, ImageDtype):
            return [data.array.get_image(i) for i in range(len(data))]
        return data.tolist()

    return data

__all__ = [
    "logger",
    "DataFrame",
    "ImageDtype", 
    "ImageArray", 
    "AudioDtype", 
    "AudioArray", 
    "FileDtype", 
    "FileArray",
    "convert_to_base_data", 
    "configure_llm_backbone"
]



"""
DataFrame class for handling data in a tabular format.
To begin, DataFrame is a combination of regular pandas DataFrame and additional unstructured data types (e.g., text, images, and audio).
Example:
| location | house price | comment | house picture |
|----------|-------------|---------|---------------|
|161 Auburn St. Unit 161, Cambridge, MA 02139| $1,000,000 | "Great location!" | ![house](house.jpg) (a link to the house image) |
|...|...|...|...|

A simple version that directly uses pandas DataFrame and supports image data type through pandas api extension, like lotus.
Later, we will implement an independent DataFrame of pandas DataFrame. The usage will be like:
```python
import mahjong as mjg

df = mjg.DataFrame({
    'location': ['161 Auburn St. Unit 161, Cambridge, MA 02139'],
    'house price': ['$1,000,000'],
    'comment': ['Great location!'],
    'house picture': [mjg.Image('house.jpg')]
})
```
Moreover, in the future, we consider reading data from data lake storages (e.g, S3, Delta Lake, etc.)
"""
from typing import Callable, Literal
import pandas as pd

from nirvana.lineage.mixin import LineageMixin
from nirvana.dataframe.arrays.utils import infer_and_convert_dtype


class DataFrame(LineageMixin):
    def __init__(
        self,
        data: dict | pd.DataFrame = None,
        *args,
        **kwargs
    ):
        self._data = pd.DataFrame(data) if isinstance(data, dict) else data
        for col in self._data.columns:
            converted_array, inferred_dtype = infer_and_convert_dtype(self._data[col])
            self._data[col] = converted_array
        self.initialize()

    def __len__(self):
        _len = self.nrows
        return _len
    
    def __contains__(self, item):
        return self.columns.__contains__(item)
    
    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(nrows={self.nrows}, ncols={len(self.columns)})"
    
    @property
    def columns(self):
        return list(self._data.columns)
    
    @columns.setter
    def columns(self, columns):
        self._data.columns = columns
    
    @property
    def primary_key(self):
        return self._primary_key
    
    @property
    def nrows(self):
        return self._data.shape[0]
    
    @classmethod
    def from_external_file(cls, path: str, sep=',', **kwargs):
        df = pd.read_table(path, sep=sep, **kwargs)
        return cls(df)
    
    def _get(self, posidx, materialize: bool = False):
        if isinstance(posidx, str):
            # str index => column selection (AbstractColumn)
            if posidx in self.columns:
                return self._data[posidx]
            raise KeyError(f"Column `{posidx}` does not exist.")

        elif isinstance(posidx, int):
            # int index => single row
            return self._data.iloc[posidx]
        
        index_type = None
        if isinstance(posidx, slice):
            index_type = "row"
        elif (isinstance(posidx, tuple) or isinstance(posidx, list)) and len(posidx):
            if isinstance(posidx[0], str):
                index_type = "column"
            else:
                index_type = "row"
        else:
            raise TypeError("Invalid index type: {}".format(type(posidx)))
        
        if index_type == "row":
            return self._data.iloc[posidx]
        elif index_type == "column":
            return self._data[posidx]

    def __getitem__(self, posidx):
        return self._get(posidx)

    def semantic_map(
        self,
        user_instruction: str,
        input_columns: list[str],
        output_columns: list[str],
        context: list[dict] | str | None = None,
        model: str | None = None,
        func: Callable | None = None,
        strategy: Literal["plain", "fewshot", "self-refine"] = "plain",
        limit: int | None = None,
        rate_limit: int = 16,
        assertions: list[Callable] | None = [],
    ):
        op_kwargs = {
            "user_instruction": user_instruction,
            "input_columns": input_columns,
            "output_columns": output_columns,
            "context": context,
            "model": model,
            "tool": func,
            "strategy": strategy,
            "limit": limit,
            "rate_limit": rate_limit,
            "assertions": assertions,
        }
        data_kwargs = {
            "left_input_fields": self.leaf_node.node_fields.output_fields,
            "right_input_fields": [],
            "output_fields": self.leaf_node.node_fields.left_input_fields + output_columns,
        }
        self.add_operator(op_name="map",
                          op_kwargs=op_kwargs,
                          data_kwargs=data_kwargs,
                          rate_limit=rate_limit)
        
    def semantic_filter(
        self,
        user_instruction: str,
        input_columns: list[str],
        func: Callable | None = None,
        context: list[dict] | str | None = None,
        model: str | None = None,
        strategy: Literal["plain", "fewshot", "self-refine"] = "plain",
        limit: int | None = None,
        rate_limit: int = 16,
        assertions: list[Callable] | None = [],
    ):
        op_kwargs = {
            "user_instruction": user_instruction,
            "input_columns": input_columns,
            "context": context,
            "model": model,
            "tool": func,
            "strategy": strategy,
            "limit": limit,
            "rate_limit": rate_limit,
            "assertions": assertions,
        }
        data_kwargs = {
            "left_input_fields": self.leaf_node.node_fields.output_fields,
            "right_input_fields": [],
            "output_fields": self.leaf_node.node_fields.output_fields,
        }
        self.add_operator(op_name="filter",
                          op_kwargs=op_kwargs,
                          data_kwargs=data_kwargs,
                          rate_limit=rate_limit)
        
    def semantic_reduce(
        self,
        user_instruction: str,
        input_column: str,
        context: list[dict] | str | None = None,
        model: str | None = None,
        func: Callable | None = None,
        strategy: Literal["plain"] = "plain",
        rate_limit: int = 16,
        assertions: list[Callable] | None = [],
    ):
        op_kwargs = {
            "user_instruction": user_instruction,
            "input_columns": [input_column],
            "context": context,
            "model": model,
            "tool": func,
            "strategy": strategy,
            "rate_limit": rate_limit,
            "assertions": assertions,
        }
        data_kwargs = {
            "left_input_fields": self.leaf_node.node_fields.output_fields,
            "right_input_fields": [],
            "output_fields": []
        }
        self.add_operator(op_name="reduce",
                          op_kwargs=op_kwargs,
                          data_kwargs=data_kwargs,
                          rate_limit=rate_limit)
        
    def semantic_join(
        self,
        other: "DataFrame",
        user_instruction: str,
        left_on: str,
        right_on: str,
        how: Literal["inner", "left", "right"] = "inner",
        context: list[dict] | str | None = None,
        model: str | None = None,
        func: Callable | None = None,
        strategy: Literal["nest", "block"] = "nest",
        limit: int | None = None,
        rate_limit: int = 16,
        assertions: list[Callable] | None = [],
        batch_size: int = 5,
    ):
        union_fields = (
            list(set(self.leaf_node.node_fields.output_fields) | set(other.leaf_node.node_fields.output_fields))
        )
        op_kwargs = {
            "user_instruction": user_instruction,
            "left_on": [left_on],
            "right_on": [right_on],
            "how": how,
            "context": context,
            "model": model,
            "tool": func,
            "strategy": strategy,
            "limit": limit,
            "rate_limit": rate_limit,
            "assertions": assertions,
            "batch_size": batch_size,
        }
        data_kwargs = {
            "input_left_fields": self.leaf_node.node_fields.output_fields,
            "input_right_fields": other.leaf_node.node_fields.output_fields,
            "output_fields": union_fields
        }
        self.add_operator(op_name="join",
                          op_kwargs=op_kwargs,
                          data_kwargs=data_kwargs,
                          other=other,
                          rate_limit=rate_limit)
        
    def semantic_rank(
        self,
        user_instruction: str,
        input_column: str,
        descend: bool = True,
        context: list[dict] | str | None = None,
        model: str | None = None,
        func: Callable | None = None,
        strategy: Literal["plain"] = "plain",
        limit: int | None = None,
        rate_limit: int = 16,
        assertions: list[Callable] | None = [],
    ):
        op_kwargs = {
            "user_instruction": user_instruction,
            "input_columns": [input_column],
            "descend": descend,
            "context": context,
            "model": model,
            "tool": func,
            "strategy": strategy,
            "limit": limit,
            "rate_limit": rate_limit,
            "assertions": assertions,
        }
        data_kwargs = {
            "left_input_fields": self.leaf_node.node_fields.output_fields,
            "right_input_fields": [],
            "output_fields": self.leaf_node.node_fields.output_fields,
        }
        self.add_operator(op_name="rank",
                          op_kwargs=op_kwargs,
                          data_kwargs=data_kwargs,
                          rate_limit=rate_limit)
    
    def optimize_and_execute(self, optim_config = None):
        self.create_plan_optimizer(optim_config)
        if self.optimizer.config.do_logical_optimization:
            self.leaf_node = self.optimizer.optimize_logical_plan(self.leaf_node)
        if self.optimizer.config.do_physical_optimization:
            output, cost, runtime = self.optimizer.optimize_physical_plan(
                self.leaf_node,
            )
        else:
            output, cost, runtime = self.execute()
        return output, cost, runtime



import asyncio
from dataclasses import dataclass, field
from typing import Any, Callable
from abc import ABC, abstractmethod

from nirvana.executors.llm_backbone import LLMClient
from nirvana.executors.tools import BaseTool


@dataclass
class BaseOpOutputs:
    cost: float = field(default=0.0)


class BaseOperation(ABC):
    llm: LLMClient = None

    def __init__(
        self, 
        op_name: str,
        user_instruction: str = "",
        context: list[dict] | str | None = None,
        model: str | None = None,
        tool: BaseTool | None = None,
        strategy: str | None = "plain",
        limit: int | None = None,
        rate_limit: int = 16,
        assertions: list[Callable] | None = [],
    ):
        self.op_name = op_name
        self.user_instruction = user_instruction
        self.context = context
        self.model = model if model else self.llm.default_model
        self.tool = tool
        self.strategy = "plain" if strategy is None else strategy
        self.limit = limit
        self.semaphore = asyncio.Semaphore(rate_limit)
        self.assertions = assertions

    @classmethod
    def set_llm(cls, llm: LLMClient):
        cls.llm = llm
        
    @property
    def op_kwargs(self) -> dict:
        return {
            "user_instruction": self.user_instruction,
            "context": self.context,
            "model": self.model,
            "tool": self.tool,
            "strategy": self.strategy,
            "assertions": self.assertions,
            "limit": self.limit,
            "rate_limit": self.semaphore._value,
        }

    @property
    def dependencies(self) -> list[str]:
        raise NotImplementedError("Subclasses must implement dependencies property.")

    @property
    def generated_fields(self) -> list[str]:
        raise NotImplementedError("Subclasses must implement generated_fields property.")

    def has_udf(self) -> bool:
        return callable(self.tool)

    def add_assertion(self, assertion: Callable | list[Callable]):
        if isinstance(assertion, list):
            self.assertions.extend(assertion)
        else:
            self.assertions.append(assertion)

    @abstractmethod
    def execute(self, *args, **kwargs):
        raise NotImplementedError(f"The operator {self.op_name} is not implemented.")



from nirvana.optim.rules.filter_pushdown import FilterPushdown
from nirvana.optim.rules.filter_pullup import FilterPullup
from nirvana.optim.rules.map_pullup import MapPullup
from nirvana.optim.rules.non_llm_pushdown import NonLLMPushdown
from nirvana.optim.rules.non_llm_replace import NonLLMReplace
from nirvana.optim.rules.operator_fusion import OperatorFusion


__all__ = [
    "FilterPushdown",
    "FilterPullup",
    "MapPullup",
    "NonLLMPushdown",
    "NonLLMReplace",
    "OperatorFusion",
]



import os
import asyncio
from enum import Enum
from typing import Any, List, Iterable, Union
from dataclasses import dataclass
import bm25s
import Stemmer
import numpy as np
import pandas as pd
from scipy.spatial.distance import cosine

from nirvana.ops.base import BaseOpOutputs, BaseOperation
from nirvana.ops.prompt_templates.discover_prompter import SummarizePrompter, RerankPrompter


def summarize_wrapper(**kwargs):
    pass


def discover_wrapper(**kwargs):
    pass


class DiscoveryType(Enum):
    CONTEXT = 1
    COLUMN = 2
    RECORD = 3


@dataclass
class SummarizeOpOutputs(BaseOpOutputs):
    column_summary: str = None
    record_summary: str = None


@dataclass
class VectorizeOpOutputs(BaseOpOutputs):
    doc_ids: List[str] = None
    semantic_vectors: List[List[float]] = None
    lexical_vectors: Union[List[List[str]], bm25s.tokenization.Tokenized] = None
    documents: List[str] = None


@dataclass
class DiscoverOpOutputs(BaseOpOutputs):
    table_ids: List[int] = None
    scores: List[float] = None
    types: List[DiscoveryType] = None


class SummarizeOperation(BaseOperation):
    """ TODO: apply data summary on a single data (table) 
    Operation for data summary that summarize the core concept of a table from its columns and rows.
    Currently, the main purpose of this operation is serving data discovery.
    
    The algorithm is adopted from [Pneuma: Leveraging LLMs for Tabular Data Representation and Retrieval in an End-to-End System](https://arxiv.org/abs/2504.09207)
    """
    
    def __init__(
            self,
            *args,
            **kwargs,
    ):
        super().__init__("summarize", *args, **kwargs)
        self.prompter = SummarizePrompter()

    async def _summarize_column_by_llm(self, target_column: Any, all_columns: Iterable[str], **kwargs):
        async with self.semaphore:
            columns = " | ".join(all_columns)
            full_prompt = self.prompter.generate_prompt(target_column, columns)
            output = await self.llm(full_prompt, parse_tags=False, **kwargs)
            return output["output"], output["cost"]
        
    def _summarize_record(self, sampled_rows: pd.DataFrame, **kwargs):
        rows = []
        for row_idx, row in sampled_rows.iterrows():
            row = " | ".join([f"{col}: {val}" for col, val in row.items()])
            rows.append(row.strip())
        return rows

    def merge_summarizes(self, summarizes: List[str], strategy=None, max_length=None):
        merged_summary = ""
        if strategy == "truncate":
            for desc_id, desc in enumerate(summarizes):
                desc = f"{desc} || " if desc_id < len(summarizes) - 1 else desc
                if len(merged_summary) + len(desc) <= max_length:
                    merged_summary += desc
                else:
                    break
        else:
            for desc_id, desc in enumerate(summarizes):
                desc = f"{desc} || " if desc_id < len(summarizes) - 1 else desc
                merged_summary += desc
        return merged_summary

    async def execute(
            self,
            data_from_lake: pd.DataFrame,
            column_level: bool = True,
            record_level: bool = True,
            sample_size: int = 5,
            strategy: str = None,
            max_length: int = 4096,
            **kwargs
    ):
        token_cost = 0.0
        column_summary, record_summary = None, None
        if column_level:
            tasks = []
            for col in data_from_lake.columns:
                tasks.append(
                    asyncio.create_task(self._summarize_column_by_llm(col, data_from_lake.columns, **kwargs))
                )
            results = await asyncio.gather(*tasks)
            column_summarizes = [f"{col}: {result['output'].strip()}" for col, result in zip(data_from_lake.columns, results)]
            token_cost += sum([result["cost"] for result in results])
            column_summary = self.merge_summarizes(column_summarizes, strategy, max_length)

        if record_level:
            sample_size = min(sample_size, len(data_from_lake))
            sampled_rows = data_from_lake.sample(n=sample_size).reset_index(drop=True)
            record_summarizes = self._summarize_record(sampled_rows, **kwargs)
            record_summary = self.merge_summarizes(record_summarizes, strategy, max_length)
        
        return SummarizeOpOutputs(
            column_summary=column_summary, 
            record_summary=record_summary,
            cost=token_cost
        )
    

class VectorizeOperation(BaseOperation):
    def __init__(self, *args, **kwargs):
        super().__init__("index", *args, **kwargs)

    async def _prepare_documents(self, lake_manager):
        doc_idx, documents = [], []
        for idx, metadata in enumerate(lake_manager):
            if not metadata.is_summarized:
                continue
            if metadata.column_summary is not None:
                documents.append(metadata.column_summary)
                doc_idx.append(f"{idx}_column")
            if metadata.record_summary is not None:
                documents.append(metadata.record_summary)
                doc_idx.append(f"{idx}_record")
            if metadata.context is not None:
                documents.append(metadata.context)
                doc_idx.append(f"{idx}_context")
            metadata.is_vector_indexed = True
            metadata.is_text_indexed = True
        return doc_idx, documents

    async def _embed_documents(self, documents: List[str], embedding_model: str):
        with self.semaphore:
            return await self.llm.create_embedding(documents, embedding_model)

    async def execute(
            self, 
            lake_manager,
            embedding_model: str,
            *args, 
            **kwargs
    ):
        doc_idx, documents = await self._prepare_documents(lake_manager)
        tasks = []
        for i in range(0, len(documents), 30000):
            tasks.append(
                asyncio.create_task(self._embed_documents(documents[i:i+30000], embedding_model))
            )
        results = await asyncio.gather(*tasks)
        embeddings = np.concat([result[0] for result in results], axis=0)
        token_cost = sum([result[1] for result in results])

        tokenized_docs = bm25s.tokenize(
            documents,
            stopwords="en",
            stemmer=Stemmer.Stemmer("english"),
            show_progress=False
        )
        return VectorizeOpOutputs(
            doc_ids=doc_idx, 
            documents=documents,
            semantic_vectors=embeddings.tolist(),
            lexical_vectors=tokenized_docs,
            cost=token_cost
        )


class DiscoverOperation(BaseOperation):
    """ TODO: apply data discovery over a data lake
    Operation for data discovery that identify relevant data(tables) from data lakes for a data analytics task (in NL).
    
    The algorithm is adopted from [Pneuma: Leveraging LLMs for Tabular Data Representation and Retrieval in an End-to-End System](https://arxiv.org/abs/2504.09207)
    """
    
    def __init__(
            self,
            *args,
            **kwargs,
    ):
        super().__init__("discover", *args, **kwargs)
        self.prompter = RerankPrompter()

    def _adjust_index_search(self, search_results, misaligned_ids, reverse_vocab_dict, text_index, query):
        docs = [docs for docs in search_results[0][0]]
        scores = [score for score in search_results[1][0]]

        misaligned_docs = [
            text_index.corpus[reverse_vocab_dict[doc_id]] for doc_id in misaligned_ids
        ]
        misaligned_scores = [
            text_index.get_scores(
                bm25s.tokenization.convert_tokenized_to_string_list(query)[0]
            )[reverse_vocab_dict[doc_id]] for doc_id in misaligned_ids
        ]
        docs.extend(misaligned_docs)
        scores.extend(misaligned_scores)

        max_score, min_score = max(scores), min(scores)
        processed_results = {
            result["id"]: (
                1 if min_score == max_score else (scores[i] - min_score) / (max_score - min_score),
                result["text"]
            )
            for i, result in enumerate(docs)
        }
        return processed_results
    
    def _adjust_vector_search(self, search_results, misaligned_ids, vector_index, query):
        misaligned_results = vector_index.get_fast(
            ids=misaligned_ids, limit=len(misaligned_ids), include=["documents", "embeddings"]
        )
        search_results["ids"][0].extend(misaligned_results["ids"])
        search_results["documents"][0].extend(misaligned_results["documents"])
        search_results["distances"][0].extend(
            cosine(query, misaligned_results["embeddings"][i])
            for i in range(len(misaligned_ids))
        )
        scores = [1 - dist for dist in search_results["distances"][0]]
        dosc = search_results["documents"][0]
        ids = search_results["ids"][0]

        max_score, min_score = max(scores), min(scores)
        processed_results = {
            ids[i]: (
                1 if min_score == max_score else (scores[i] - min_score) / (max_score - min_score),
                dosc[i]
            )
            for i in range(len(scores))
        }
        return processed_results
    
    async def _judge_relevance_by_llm(self, query, doc_id, doc=None, **kwargs):
        with self.semaphore:
            if "context" in doc_id:
                full_prompt = self.prompter.generate_prompt(query, context_desc=doc)
            else:
                full_prompt = self.prompter.generate_prompt(query, columns_desc=doc)
            outputs = await self.llm(full_prompt, parse_tags=True, tags=["output"], **kwargs)
            return outputs["output"], outputs["cost"]
    
    async def _rerank_search_results(self, search_results, query, **kwargs):
        tasks = []
        for doc_id, _, doc in search_results:
            tasks.append(
                asyncio.create_task(self._judge_relevance_by_llm(query, doc_id, doc, **kwargs))
            )
        results = await asyncio.gather(*tasks)
        table_relevance = ["true" in result["output"].lower() for result in results]
        token_cost = sum([result["cost"] for result in results])
        rerank_results = (
            [result for idx, result in enumerate(search_results) if table_relevance[idx]] + 
            [result for idx, result in enumerate(search_results) if not table_relevance[idx]]
        )
        return rerank_results, token_cost
    
    def postprocess_discover_output(self, search_results, **kwargs):
        table_ids, scores, types = [], [], []
        for doc_id, score, doc in search_results:
            if "context" in doc_id:
                types.append(DiscoveryType.CONTEXT)
            elif "column" in doc_id:
                types.append(DiscoveryType.COLUMN)
            else:
                types.append(DiscoveryType.RECORD)
            table_ids.append(int(doc_id.split("_")[0]))
            scores.append(score)
        return table_ids, scores, types

    async def execute(
            self, 
            query: str,
            vector_index,
            text_index,
            embedding_model: str,
            topk: int = 1,
            pool_factor: int = 5,
            alpha: float = 0.5,
            **kwargs
    ):
        pool_size = pool_factor * topk
        token_cost = 0.0
        # process the query
        tokenized_query = bm25s.tokenize(
            query,
            stopwords="en",
            stemmer=Stemmer.Stemmer("english"),
            show_progress=False
        )
        embedded_query, embedding_query_cost = await self.llm.create_embedding(query, embedding_model)
        token_cost += embedding_query_cost

        # search relevant data with vector index and text index
        text_search_results = text_index.retrieve(
            tokenized_query,
            k=pool_size,
            show_progress=False
        )
        vector_search_results = vector_index.query(
            query_embeddings=embedded_query.tolist(),
            n_results=pool_size
        )

        # align search results between vector index and text index
        doc_ids_from_vector_search = set([doc_id for doc_id in vector_search_results["ids"][0]])
        doc_ids_from_text_search = set([result["id"] for result in text_search_results[0][0]])
        reverse_vocab_dict = {
            docs["id"]: idx for idx, docs in enumerate(text_index.corpus)
        }
        adjusted_text_search_results = self._adjust_index_search(
            text_search_results, 
            list(doc_ids_from_vector_search - doc_ids_from_text_search),
            reverse_vocab_dict,
            text_index,
            tokenized_query
        )
        adjusted_vector_search_results = self._adjust_vector_search(
            vector_search_results, 
            list(doc_ids_from_text_search - doc_ids_from_vector_search),
            vector_index,
            embedded_query
        )

        # rerank the search results and generate hybrid search results
        hybrid_search_results = []
        for doc_id in sorted(doc_ids_from_vector_search | doc_ids_from_text_search):
            result_from_text_search = adjusted_text_search_results.get(doc_id)
            result_from_vector_search = adjusted_vector_search_results.get(doc_id)
            combined_score = alpha * result_from_text_search[0] + (1 - alpha) * result_from_vector_search[0]
            if result_from_text_search is None:
                doc = result_from_vector_search[1]
            else:
                doc = result_from_text_search[1]
            hybrid_search_results.append((doc_id, combined_score, doc))
        
        sorted_hybrid_results = sorted(hybrid_search_results, key=lambda x: (-x[1], x[0]))[:topk]
        reranked_search_results, rerank_cost = await self._rerank_search_results(
            sorted_hybrid_results, 
            query, 
            **kwargs
        )
        token_cost += rerank_cost

        table_ids, scores, types = self.postprocess_discover_output(reranked_search_results)
        return DiscoverOpOutputs(table_ids=table_ids, scores=scores, types=types, cost=token_cost)
